from moai.utils.arguments import assert_choices
from moai.monads.utils.spatial import spatial_dims
from collections import namedtuple

import torch
import functools
import logging

log = logging.getLogger(__name__)

__all__ = [
    "Interpolate",
    "BilinearDownsample_x2",
    "NearestDownsample_x2",
    "BilinearDownsample",
    "NearestDownsample",
]

#NOTE: align corners doc: https://discuss.pytorch.org/t/what-we-should-use-align-corners-false/22663/12

class Interpolate(torch.nn.Module):
    __MODES__ = ['nearest', 'linear', 'bilinear', 'area', 'bicubic', 'trilinear']

    STATIC_PARAMS = namedtuple('params', ['mode', 'align_corners', 'recompute_scale_factor'])

    def __init__(self,
        scale:                  float=0.0,
        width:                  int=-1,
        height:                 int=-1,
        #TODO: support depth as well for volumetric interpolation
        mode:                   str='bilinear', # one of ['nearest', 'linear', 'bilinear', 'area', 'bicubic', 'trilinear']            
        align_corners:          bool=False,
        recompute_scale_factor: bool=False,
    ):
        super(Interpolate, self).__init__()
        assert_choices(log, "interpolation mode", mode, Interpolate.__MODES__)
        if not scale > 0.0 and not (width > 0 and height > 0):
            log.error(f"Either scale ({scale}) or dimensions ([{width} x {height}]) need to be set for Interpolate.")
        self.func = functools.partial(
            torch.nn.functional.interpolate,
            scale_factor=scale, mode=mode, 
            align_corners=align_corners if mode != 'nearest' and mode != 'area' else None,
            recompute_scale_factor=recompute_scale_factor,
        ) if scale > 0.0 and width < 0 and height < 0 else\
            functools.partial(
                torch.nn.functional.interpolate,
                size=(height, width), mode=mode, 
                align_corners=align_corners if mode != 'nearest' and mode != 'area' else None,
                recompute_scale_factor=recompute_scale_factor,
            )
        self.params = Interpolate.STATIC_PARAMS(mode, align_corners, recompute_scale_factor)

    def forward(self,
        image:  torch.Tensor,
        target: torch.Tensor=None,
    ) -> torch.Tensor:
        return self.func(image) if target is None else torch.nn.functional.interpolate(
            image, size=target.shape[2:], **self.params._asdict()
        )#TODO: create a common func for spatial dims as tuple/list

BilinearDownsample_x2 = functools.partial(Interpolate,
    mode='bilinear', scale=0.5,
)
NearestDownsample_x2 = functools.partial(Interpolate,
    mode='nearest', scale=0.5,
)
BilinearSampling = functools.partial(Interpolate,
    mode='bilinear',
)
NearestSampling = functools.partial(Interpolate,
    mode='nearest',
)


from moai.utils.torch import get_submodule

import torch
import toolz
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["ZeroParams"]

class ZeroParams(typing.Callable[[torch.nn.Module], None]):
    def __init__(self, 
        keys:           typing.Sequence[str],
    ):
        self.keys = keys
        
    def __call__(self,
        module: torch.nn.Module
    ) -> None:        
        for key in self.keys:
            try:
                # m = module.get_submodule(key) #NOTE: #TODO: for PyTorch 1.10
                # split = key.split('.')
                # def _getattr(object: typing.Any, key: str):
                #     return getattr(object, key, None)
                # m = toolz.reduce(_getattr, split, module)
                m = get_submodule(module, key)
                if m is not None:
                    log.info(f"Zeroing out parameter: {key}.")
                    with torch.no_grad(): #TODO: remove this and add in root apply call
                        m.zero_()
            except:
                break
            



import math
import torch
import typing

__all__ = ['AdamP']

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/adamp.py

class AdamP(torch.optim.Optimizer):
    r"""Implements AdamP algorithm.

    - **Paper**: [Slowing Down the Weight Norm Increase in Momentum-based Optimizers](https://arxiv.org/pdf/2006.08217.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)    

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: learning rate (default: 1e-3)
        betas: coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps: term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay: weight decay (L2 penalty) (default: 0)
        delta: threhold that determines whether a set of parameters is scale
            invariant or not (default: 0.1)
        wd_ratio: relative weight decay applied on scale-invariant parameters
            compared to that applied on scale-variant parameters (default: 0.1)
        nesterov: enables Nesterov momentum (default: False)

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.AdamP(model.parameters(), lr=0.1)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

     __ https://arxiv.org/abs/2006.08217

    Note:
        Reference code: https://github.com/clovaai/AdamP
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=0,
        delta: float=0.1,
        wd_ratio: float=0.1,
        nesterov: bool=False,
    ) -> None:
        if lr <= 0.0:
            raise ValueError('Invalid learning rate: {}'.format(lr))
        if eps < 0.0:
            raise ValueError('Invalid epsilon value: {}'.format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(
                'Invalid beta parameter at index 0: {}'.format(betas[0])
            )
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(
                'Invalid beta parameter at index 1: {}'.format(betas[1])
            )
        if weight_decay < 0:
            raise ValueError(
                'Invalid weight_decay value: {}'.format(weight_decay)
            )
        if delta < 0:
            raise ValueError('Invalid delta value: {}'.format(delta))
        if wd_ratio < 0:
            raise ValueError('Invalid wd_ratio value: {}'.format(wd_ratio))

        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            delta=delta,
            wd_ratio=wd_ratio,
            nesterov=nesterov,
        )
        super(AdamP, self).__init__(params, defaults)

    @staticmethod
    def _channel_view(x):
        return x.view(x.size(0), -1)

    @staticmethod
    def _layer_view(x):
        return x.view(1, -1)

    @staticmethod
    def _cosine_similarity(x, y, eps, view_func):
        x = view_func(x)
        y = view_func(y)

        x_norm = x.norm(dim=1).add_(eps)
        y_norm = y.norm(dim=1).add_(eps)
        dot = (x * y).sum(dim=1)

        return dot.abs() / x_norm / y_norm

    def _projection(self, p, grad, perturb, delta, wd_ratio, eps):
        wd = 1
        expand_size = [-1] + [1] * (len(p.shape) - 1)
        for view_func in [self._channel_view, self._layer_view]:

            cosine_sim = self._cosine_similarity(grad, p.data, eps, view_func)

            if cosine_sim.max() < delta / math.sqrt(view_func(p.data).size(1)):
                p_n = p.data / view_func(p.data).norm(dim=1).view(
                    expand_size
                ).add_(eps)
                perturb -= p_n * view_func(p_n * perturb).sum(dim=1).view(
                    expand_size
                )
                wd = wd_ratio

                return perturb, wd

        return perturb, wd

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                grad = p.grad.data
                beta1, beta2 = group['betas']
                nesterov = group['nesterov']

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)

                # Adam
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']

                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                denom = (exp_avg_sq.sqrt() / math.sqrt(bias_correction2)).add_(
                    group['eps']
                )
                step_size = group['lr'] / bias_correction1

                if nesterov:
                    perturb = (beta1 * exp_avg + (1 - beta1) * grad) / denom
                else:
                    perturb = exp_avg / denom

                # Projection
                wd_ratio = 1
                if len(p.shape) > 1:
                    perturb, wd_ratio = self._projection(
                        p,
                        grad,
                        perturb,
                        group['delta'],
                        group['wd_ratio'],
                        group['eps'],
                    )

                # Weight decay
                if group['weight_decay'] > 0:
                    p.data.mul_(
                        1 - group['lr'] * group['weight_decay'] * wd_ratio
                    )

                # Step
                p.data.add_(perturb, alpha=-step_size)

        return loss


from moai.supervision.losses.regression.robust.berhu import Berhu
from moai.supervision.losses.regression.robust.welsch import Welsch
from moai.supervision.losses.regression.robust.wing import Wing
from moai.supervision.losses.regression.robust.gm import GemanMcClure

__all__ = [
    "Berhu",
    "Welsch",
    "Wing",
    "GemanMcClure",
]


from moai.parameters.optimization.schedule.schedulers.identity import Identity
from moai.parameters.optimization.schedule.schedulers.cyclic_scheduler import CyclicLRWithRestarts

__all__ = [
    "Identity",
    "CyclicLRWithRestarts",
]


from moai.monads.generation.mesh.mesh_type import TriangleMesh
from moai.utils.arguments import assert_path

import torch
import logging

log = logging.getLogger(__name__)

class Vertices(torch.nn.Module):
    def __init__(self,
        filename: str
    ):
        super(Vertices,self).__init__()
        assert_path(log, "mesh filename", filename)
        mesh = TriangleMesh.from_obj(filename)
        vertices = mesh.vertices.unsqueeze(0)
        faces = mesh.faces.int() #TODO: support for faces too
        self.register_buffer("vertices", vertices.float())

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b = x.shape[0]
        return self.vertices.expand(b,*self.vertices.shape[1:])


import pytorch_lightning
import omegaconf.omegaconf
import hydra.utils as hyu
import logging
import functools

log = logging.getLogger(__name__)

__all__ = ["Loggers"]

class Loggers(pytorch_lightning.loggers.base.LoggerCollection):
    def __init__(self,
        loggers: omegaconf.DictConfig={},
        name: str="Sequential",
        version: int=0
    ):
        super(Loggers, self).__init__(logger_iterable=[
            hyu.instantiate(logger) for logger in loggers.values()
        ])
        self._name = name
        self._version = version

    @property
    def name(self) -> str:
        names = [logger.name for logger in self]
        return functools.reduce(lambda seed, item: f"{seed}_{item}", set(names))


import torch
import functools

__all__ = [
    "Downsample2d",
]

class Downsample2d(torch.nn.Module):
    def __init__(self,
        scale: float=0.5,
        mode: str="bilinear"
    ):
        super(Downsample2d, self).__init__()
        self.downsample = functools.partial(torch.nn.functional.interpolate,
            scale_factor=scale,
            mode=mode
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.downsample(x)


from moai.monads.utils.spatial import spatial_dim_list

import moai.nn.convolution as mic
import moai.nn.residual as mires
import moai.nn.deconvolution as midec 
import moai.nn.utils as miu
import moai.nn.sampling.spatial.downsample as mids
import moai.nn.sampling.spatial.upsample as mius

import torch
import typing
import functools
import omegaconf.omegaconf

__all__ = ["TopBranch", "AllBranches","Higher"]

class TopBranch(torch.nn.Module):
    def __init__(self, 
        stages:         int,
        start_features: int,
        out_features:   int,
        convolution:    str="conv2d",
        activation:     str="none",
        kernel_size:    int=1,
        padding:        int=1,
        inplace:        bool=True
    ):
        super(TopBranch, self).__init__()
        self.conv = mic.make_conv_block(
            block_type="conv2d",
            convolution_type=convolution,
            in_features=start_features,
            out_features=out_features,
            activation_type=activation,
            convolution_params={
                'kernel_size': kernel_size,
                'padding': padding                
            },
            activation_params={
                'inplace': inplace
            }
        )
        self.sigmoid = torch.nn.Sigmoid()

    def forward(self, 
        all_branches: typing.Union[typing.Tuple[torch.Tensor, ...], typing.List[torch.Tensor]]
    ) -> torch.Tensor:
        return self.conv(all_branches[0])

class AllBranches(torch.nn.Module):
    def __init__(self, 
        stages:         int,
        start_features: int,
        out_features:   int,
        projection:     omegaconf.DictConfig,
        prediction:     omegaconf.DictConfig,
        upsample:       omegaconf.DictConfig
    ):
        super(AllBranches, self).__init__()
        in_features = sum((start_features * (2 ** i) for i in range(stages)))
        self.projection = mic.make_conv_block(
            block_type='conv2d',
            convolution_type=projection.convolution,
            in_features=in_features,
            out_features=in_features,
            activation_type=projection.activation,
            convolution_params={
                'kernel_size': projection.kernel_size,
                'padding': projection.padding,
            },
            activation_params={
                'inplace': projection.inplace
            }
        )
        self.prediction = mic.make_conv_block(
            block_type='conv2d',
            convolution_type=prediction.convolution,
            in_features=in_features,
            out_features=out_features,
            activation_type=prediction.activation,
            convolution_params={
                'kernel_size': prediction.kernel_size,
                'padding': prediction.padding,
            },
            activation_params={
                'inplace': prediction.inplace
            }
        )
        self.upsample = functools.partial(torch.nn.functional.interpolate,
            mode=upsample.mode
        )
        
    def forward(self, 
        all_branches: typing.Union[typing.Tuple[torch.Tensor, ...], typing.List[torch.Tensor]]
    ) -> torch.Tensor:
        top = all_branches[0]
        dims = spatial_dim_list(top)
        up = (self.upsample(x, size=dims) for x in all_branches[1:])
        concat = torch.cat([top, *up], dim=1)
        return self.prediction(self.projection(concat))


class Higher(torch.nn.Module):
    def __init__(
        self,
        stages: int,
        start_features: int, #in channels
        out_features: int, #initial out channels
        deconvolution_modules: int, #nr of deconvolution modules
        deconvolution: omegaconf.DictConfig,
        residual:   omegaconf.DictConfig,
        final: omegaconf.DictConfig,
        aggregator: omegaconf.DictConfig,
    ):
        super(Higher, self).__init__()

        self.deconvolution_modules = deconvolution_modules
        self.deconvolution = deconvolution

        start_features_ = lambda i : start_features if i == 0 else deconvolution.deconv_out_features
        self.deconv_layers = torch.nn.ModuleList([
            torch.nn.Sequential(*[
                midec.make_deconv_block(
                    block_type = deconvolution.block,
                    in_features = (start_features_(i) + out_features) \
                                 if deconvolution.concat \
                                 else start_features_(i),
                    out_features=deconvolution.deconv_out_features,
                    deconvolution_type = deconvolution.type,
                    activation_type=deconvolution.activation,
                    deconvolution_params={
                        'kernel_size': deconvolution.kernel_size,
                        'padding': deconvolution.padding,
                        'output_padding': deconvolution.output_padding,
                        'stride': 2,
                    }),
                torch.nn.Sequential(*[
                    mires.make_residual_block(
                        block_type=residual.type,
                        convolution_type=residual.convolution,
                        in_features=deconvolution.deconv_out_features,
                        out_features=deconvolution.deconv_out_features,
                        bottleneck_features=residual.bottleneck_features,
                        activation_type=residual.activation,
                        strided=False,
                        convolution_params={
                                'bias':         False,
                        }) for r in range(deconvolution.residual_units)
                ])
            ]) for i in range(deconvolution_modules)
        ])

        #final layer
        self.final_layers = torch.nn.ModuleList([
            mic.make_conv_block(
                block_type="conv2d",
                convolution_type=final.convolution,
                in_features=start_features,
                out_features=out_features,
                activation_type=final.activation,
                convolution_params={
                    'kernel_size': final.kernel_size,
                    'padding': final.padding,
                }
            ),
            torch.nn.ModuleList([
                mic.make_conv_block(
                    block_type="conv2d",
                    convolution_type=final.convolution,
                    in_features=deconvolution.deconv_out_features,
                    out_features=out_features,
                    activation_type=final.activation,
                    convolution_params={
                        'kernel_size': final.kernel_size,
                        'padding': final.padding,
                    }
                ) for i in range(deconvolution_modules)
            ])
        ])

        self.aggregator_func = functools.partial(
            torch.nn.functional.interpolate,
            size=(aggregator.height,aggregator.width),
            mode=aggregator.mode,
            align_corners=aggregator.align_corners,
            recompute_scale_factor = aggregator.recompute_scale_factor
        )
    
    def forward(self, 
        all_branches: typing.Union[typing.Tuple[torch.Tensor, ...], typing.List[torch.Tensor]]
    ) -> torch.Tensor:
        
        isTraining = self.final_layers[0].training
        x = all_branches[0]
        y = self.final_layers[0](x) #output from top branch
        final_outputs = []
        if isTraining:
            final_outputs.append(y)
        else:
            heatmaps_avg = []
            y_upscaled = self.aggregator_func(y)
            heatmaps_avg.append(y_upscaled)
        for i in range(self.deconvolution_modules):
            if self.deconvolution.concat:
                x = torch.cat((x, y), 1)
            x = self.deconv_layers[i](x)
            y = self.final_layers[1][i](x)
            if isTraining:
                final_outputs.append(y)
            else:
                #upscale output to be used in the aggregator
                y_upscaled = self.aggregator_func(y)
                heatmaps_avg.append(y_upscaled)
        
        if not isTraining:
            final_outputs.append(
                torch.mean(torch.stack(heatmaps_avg), dim = 0)
            )
            final_outputs.append(torch.zeros((1)))
        
        return final_outputs


from moai.utils.arguments import (
    assert_choices,
    assert_sequence_size
)

import torch
import functools
import itertools
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["FiniteDifference"]

class FiniteDifference(torch.nn.Module): #TODO: support for order
    
    class ReflectionPad3d(torch.nn.modules.padding._ReflectionPadNd):
        def __init__(self, padding: typing.Sequence[int]):
            super(FiniteDifference.ReflectionPad3d, self).__init__()
            self.padding = torch.nn.modules.utils._ntuple(6)(padding)

    __PAD_MAP__ = {
        "reflect": {
            1: torch.nn.ReflectionPad1d,
            2: torch.nn.ReflectionPad2d,
            3: FiniteDifference.ReflectionPad3d
        },
        "replicate": {
            1: torch.nn.ReplicationPad1d,
            2: torch.nn.ReplicationPad2d,
            3: torch.nn.ReplicationPad3d,
        },
        "zero": {
            1: functools.partial(torch.nn.ConstantPad1d, value=0.0),
            2: torch.nn.ZeroPad2d,
            3: functools.partial(torch.nn.ConstantPad3d, value=0.0),
        }
    }

    __PAD_TYPES__ = ['reflect', 'replicate', 'zero']
    __MODES__ = ['central', 'forward', 'backward']
    __REDUCTIONS__ = ['none', 'magnitude', 'absum']

    def __init__(self,
        dims:               typing.Sequence[int],
        mode:               str='central', # see __MODES__
        pad:                str='reflect', # see __PAD_TYPES__
        reduction:          str='magnitude', # see __REDUCTIONS__
        keep_size:          bool=False,
    ):
        super(FiniteDifference, self).__init__()
        assert_choices(log, "padding type", pad, FiniteDifference.__PAD_TYPES__)
        assert_choices(log, "mode", mode, FiniteDifference.__MODES__)
        assert_sequence_size(log, "dimensions", dims, max_size=3)
        self.min_dim = min(dims)
        self.scale = 0.5 if mode == 'central' else 1.0
        self.pad = self.get_padding(dims, pad, mode, keep_size)
        size_offset = 2 if mode == 'central' else 1
        nele_func = (lambda d: d) if keep_size else (lambda d: d - size_offset)
        self.nelements_func = lambda t: [nele_func(d) for d in t.shape]
        self.lhs_pad = torch.nn.Identity() if mode == 'backward' and not keep_size else self.pad
        self.rhs_pad = torch.nn.Identity() if mode == 'forward' and not keep_size else self.pad
        lhs_inds = 2 if mode == 'central' else 1
        other_ind = 0 if mode == 'forward' else 1
        self.lhs_narrow = [functools.partial(self.get_narrower, 
            dimensions=dims, dim_only=keep_size,
            this_index=lhs_inds, other_index=other_ind, dim=d,
        ) for d in dims]
        self.rhs_narrow = [functools.partial(self.get_narrower, 
            dimensions=dims, dim_only=keep_size,
            this_index=0, other_index=other_ind, dim=d,
        ) for d in dims]
        def abs_sum(t: torch.Tensor, dim: int) -> torch.Tensor:
            return torch.sum(t.abs(), dim=[dim])
        self.reduction_func = functools.partial(abs_sum, dim=self.min_dim)\
            if reduction == 'abssum' else (
                functools.partial(torch.linalg.norm, p=2, dim=self.min_dim)
                if reduction == 'magnitude' else lambda t: t
            )

    def get_padding(self, 
        dims:               typing.Sequence[int],
        pad_type:           str,
        mode:               str,
        keep_size:          bool,
    ) -> torch.nn.Module:
        base_pad = [1, 1] if mode == 'central' else [0, 1]
        base_pad = list(reversed(base_pad)) if mode == 'backward' else base_pad
        padding = functools.reduce(lambda s, t: t + s, itertools.repeat(base_pad, len(dims)))
        pad_type = "ReplicationPad" # "ReflectionPad"
        pad = FiniteDifference.__PAD_MAP__[pad_type][len(dims)]\
            if keep_size else torch.nn.Identity
        return pad(padding)

    @staticmethod
    def get_narrower(
        tensor:             torch.Tensor,
        nelements:          typing.Sequence[int],
        dim:                int,
        this_index:         int,
        other_index:        int,        
        dimensions:         int,
        dim_only:           bool,
    ) -> typing.Callable:
        return functools.reduce(
            lambda t, i: torch.narrow(
                t, i, this_index if i == dim else other_index,
                nelements[i]
            ), dimensions, tensor
        )

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        lhs_padded = self.lhs_pad(tensor)
        rhs_padded = self.rhs_pad(tensor)
        nelements = self.nelements_func(tensor)
        dTds = torch.stack([
            torch.sub(
                lhs(lhs_padded, nelements), rhs(rhs_padded, nelements)
            ) for lhs, rhs in zip(self.lhs_narrow, self.rhs_narrow)
        ], dim=self.min_dim)
        if self.scale != 1.0:
            dTds = self.scale * dTds
        return self.reduction_func(dTds)



from moai.supervision.losses.distribution.kl_divergence import KL

import torch
import functools

__all__ = ["Lambda", "JS"]

class Lambda(KL):
    def __init__(self,
        lamda: float=0.5, # default lambda for Jensen-Shannon Divergence
        is_input_log:   bool=False,
        is_target_log:  bool=False,
        epsilon: float=1e-24,
    ):
        super(Lambda, self).__init__(True, True)
        self.lamda = lamda
        self.is_input_log_ = is_input_log
        self.is_target_log_ = is_target_log
        self.epsilon = epsilon

    def forward(self,
        pred: torch.Tensor,
        gt: torch.Tensor,
    ) -> torch.Tensor:
        m = (
            self.lamda * (pred.exp() if self.is_input_log_ else pred)
            + 
            (1.0 - self.lamda) * (gt.exp() if self.is_target_log_ else gt)
        )
        m = m.log()
        p = pred if self.is_input_log_ else (pred + self.epsilon).log()
        g = gt if self.is_target_log_ else (gt + self.epsilon).log()
        
        pred_to_m = super(Lambda, self).forward(p, m)
        gt_to_m = super(Lambda, self).forward(g, m)
        lamda_divergence = self.lamda * pred_to_m + (1.0 - self.lamda) * gt_to_m
        return lamda_divergence

JS = functools.partial(Lambda, lamda=0.5)


from moai.monads.utils.common import dim_list

import torch

class RMSLE(torch.nn.Module):
    def __init__(self,
        base: str='natural' # 'natural' or 'ten'
    ):
        super(RMSLE, self).__init__()
        self.log = torch.log if base == 'natural' else torch.log10

    def forward(self,
        gt:         torch.Tensor,
        pred:       torch.Tensor,
        weights:    torch.Tensor=None,
        mask:       torch.Tensor=None,
    ) -> torch.Tensor:
        pred_fix = torch.where(pred == 0.0,
            pred + 1e-24,
            pred
        )
        log_diff_sq = (self.log(gt) - self.log(pred_fix)) ** 2
        if weights is not None:
            log_diff_sq = log_diff_sq * weights
        if mask is not None:
            log_diff_sq = log_diff_sq[mask]
        if weights is None:
            return torch.mean(torch.sqrt(torch.mean(log_diff_sq, dim=dim_list(gt))))
        else:
            log_diff_sq_sum = torch.sum(log_diff_sq, dim=dim_list(gt))
            log_diff_w_sum = torch.sum(weights, dim=dim_list(gt)) # + 1e-18
            return torch.mean(torch.sqrt(log_diff_sq_sum / log_diff_w_sum))


import torch

__all__ = ["Identity"]

class Identity(torch.optim.lr_scheduler.LambdaLR):
    def __init__(self, optimizer, **kwargs):
        unity = lambda epoch: 1
        super(Identity, self).__init__(optimizer, unity)

    def __str__(self):
        return "Identity LR scheduler (i.e. no scheduling/LR changes)"


import torch
import typing
import roma
import numpy as np

__all__ = ['Rotation3D']

def _create_rotation_x(roll: float):    
    return torch.tensor([
        [1.0,       0.0,            0.0],
        [0.0,       np.cos(roll),   -np.sin(roll)],
        [0.0,       np.sin(roll),   np.cos(roll)]
    ], dtype=torch.float32)

def _create_rotation_y(pitch: float):    
    return torch.tensor([
        [np.cos(pitch),  0.0,        np.sin(pitch)],
        [0,             1.0,        0.0],
        [-np.sin(pitch), 0,          np.cos(pitch)]
    ], dtype=torch.float32)

def _create_rotation_z(yaw: float):    
    return torch.tensor([
        [np.cos(yaw),   -np.sin(yaw),   0.0],
        [np.sin(yaw),   np.cos(yaw),    0.0],
        [0.0,           0.0,            1.0]
    ], dtype=torch.float32)

class Rotation3D(torch.nn.Module):

    __CONSTRUCTORS__ = {
        'x': _create_rotation_x,
        'y': _create_rotation_y,
        'z': _create_rotation_z,
        'X': _create_rotation_x,
        'Y': _create_rotation_y,
        'Z': _create_rotation_z,
        'roll': _create_rotation_x,
        'pitch': _create_rotation_y,
        'yaw': _create_rotation_z,
    }

    def __init__(self,
        rotations:  typing.Union[str, typing.Sequence[str]], # angle@axis format 
        epsilon:    float=1e-12,
    ) -> None:
        super().__init__()
        matrices = []
        for rot in rotations:
            euler, axis = rot.split('@')
            matrices.append(Rotation3D.__CONSTRUCTORS__[axis](
                np.deg2rad(float(euler)))
            )
        self.register_buffer('rotation', 
            roma.rotmat_composition(matrices, normalize=False)
        )
        self.rotation[self.rotation.abs() < epsilon] = 0.0

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return self.rotation.expand(tensor.shape[0], 3, 3)


import moai.utils.engine as mieng

import omegaconf.omegaconf
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["Single"]

class Single(mieng.Collection):
    def __init__(self,
        items: omegaconf.DictConfig,
        arguments: typing.Sequence[typing.Any]=None,
        name: str="items",
    ):
        super(Single, self).__init__(
            items=omegaconf.DictConfig(
                {next(iter(items.keys())): next(iter(items.values()))}
            ), name=name,
            arguments=[next(iter(arguments))] if arguments else None
        )
        if len(items.values()) > 1:
            log.warning("Multiple items have been defined,"
                "but a single item scheme is used.")


from moai.visualization.single import Visualizer as Single
from moai.visualization.collection import Visualizers as Collection
from moai.visualization.latent_collection import LatentVisualizers as LatentCollection

__all__ = [
    "Single",
    "Collection",
    "LatentCollection"
]


import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing

__all__ = ["Indexed"]

class Indexed(torch.utils.data.Dataset):
    def __init__(self,
        datasets: omegaconf.DictConfig,
        augmentation: omegaconf.DictConfig=None,
    ):
        super(Indexed, self).__init__()
        if augmentation is not None:
            self.dataset = hyu.instantiate(
                augmentation,
                hyu.instantiate(next(iter(datasets.values())))
            )
        else:
            from moai.data.augmentation import NoOp
            self.dataset = NoOp(hyu.instantiate(next(iter(datasets.values()))))

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        return self.dataset[index]


from pytorch_lightning.callbacks import Callback
from pytorch_lightning.utilities.cloud_io import load as pl_load
from pytorch_lightning.utilities.upgrade_checkpoint import KEYS_MAPPING as DEPRECATED_CHECKPOINT_KEYS
from pytorch_lightning.utilities import AMPType, rank_zero_warn
from pytorch_lightning.utilities.exceptions import MisconfigurationException

try:
    from apex import amp
except ImportError:
    amp = None

import typing
import torch
import logging

log = logging.getLogger(__name__)

__all__ = ["Continued"]

class Continued(Callback):
    def __init__(self,
        filename:           str,
    ):
        super(Continued, self).__init__()
        self.filename = filename
    
    def on_fit_start(self, trainer, pl_module):        
        checkpoint = pl_load(self.filename,
            map_location=lambda storage, loc: storage
        )
        if 'optimizer_states' not in checkpoint or 'lr_schedulers' not in checkpoint:
            error = 'Trying to restore training state but checkpoint contains only the model.'
            ' This is probably due to `ModelCheckpoint.save_weights_only` being set to `True`.'
            # raise KeyError(error)
            log.error(error)

        if any([key in checkpoint for key in DEPRECATED_CHECKPOINT_KEYS]):
            error = "The checkpoint you're attempting to load follows an"
            " outdated schema. You can upgrade to the current schema by running"
            " `python -m pytorch_lightning.utilities.upgrade_checkpoint --file model.ckpt`"
            " where `model.ckpt` is your checkpoint file."
            log.error(error)            
            raise ValueError(error)

        # restore amp scaling
        if trainer.amp_backend == AMPType.NATIVE and 'native_amp_scaling_state' in checkpoint:
            trainer.scaler.load_state_dict(checkpoint['native_amp_scaling_state'])
        elif trainer.amp_backend == AMPType.APEX and 'amp_scaling_state' in checkpoint:
            amp.load_state_dict(checkpoint['amp_scaling_state'])

        trainer.global_step = checkpoint['global_step']
        trainer.current_epoch = checkpoint['epoch']

        # crash if max_epochs is lower then the current epoch from the checkpoint
        if trainer.current_epoch > trainer.max_epochs:
            m = f"""
            you restored a checkpoint with current_epoch={trainer.current_epoch}
            but the Trainer(max_epochs={trainer.max_epochs})
            """
            # raise MisconfigurationException(m)
            log.warning(m)

        # Division deals with global step stepping once per accumulated batch
        # Inequality deals with different global step for odd vs even num_training_batches
        n_accum = 1 if trainer.accumulate_grad_batches is None else trainer.accumulate_grad_batches
        expected_steps = trainer.num_training_batches / n_accum
        if trainer.num_training_batches != 0 and trainer.global_step % expected_steps > 1:
            msg = "You're resuming from a checkpoint that ended mid-epoch. "
            "This can cause unreliable results if further training is done, "
            "consider using an end of epoch checkpoint. "
            log.warning(msg)

        log.info(f"Loading optimizer state from {self.filename} @ epoch {trainer.current_epoch} & step {trainer.global_step}.")
        optimizer_states = checkpoint['optimizer_states']
        for optimizer, opt_state in zip(trainer.optimizers, optimizer_states):
            optimizer.load_state_dict(opt_state)

            # move optimizer to GPU 1 weight at a time
            # avoids OOM
            if trainer.root_gpu is not None:
                for state in optimizer.state.values():
                    for k, v in state.items():
                        if isinstance(v, torch.Tensor):
                            state[k] = v.cuda(trainer.root_gpu)

        # restore the lr schedulers
        lr_schedulers = checkpoint['lr_schedulers']
        for scheduler, lrs_state in zip(trainer.lr_schedulers, lr_schedulers):
            scheduler['scheduler'].load_state_dict(lrs_state)


from moai.utils.arguments import ensure_string_list

import moai.networks.lightning as minet
import moai.nn.convolution as mic
import moai.nn.residual as mires
import moai.nn.sampling.spatial.downsample as mids
import moai.modules.lightning as mimod
import moai.nn.utils as miu

import torch

import hydra.utils as hyu
import omegaconf.omegaconf as omegaconf
import typing
import logging

log = logging.getLogger(__name__)

#NOTE: from https://github.com/HRNet/HRNet-Bottom-Up-Pose-Estimation/blob/master/lib/models/pose_hrnet.py
#NOTE: from https://arxiv.org/pdf/1908.07919.pdf

__all__ = ["HRNet"]

class HRNet(minet.FeedForward):
    def __init__(self,
        configuration:  omegaconf.DictConfig,
        modules:        omegaconf.DictConfig,
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,
        monads:         omegaconf.DictConfig=None,
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(HRNet, self).__init__(
            data=data, parameters=parameters,
            feedforward=feedforward, monads=monads,
            supervision=supervision, validation=validation,
            export=export, visualization=visualization,            
        )
        preproc = configuration.preproc
        #NOTE: preproc = stem + layer1
        preproc_convs = []
        prev_features = configuration.in_features
        stem = preproc.stem
        for b, c, a, f, k, s, p in zip(
            stem.blocks, stem.convolutions,
            stem.activations, stem.features,
            stem.kernel_sizes, stem.strides, stem.paddings):
            preproc_convs.append(mic.make_conv_block(
                block_type=b,
                convolution_type=c,
                in_features=prev_features, 
                out_features=f,
                activation_type=a,
                convolution_params={
                    "kernel_size": k,
                    "stride": s,
                    "padding": p,
                },
            ))
            prev_features = f
        residual = preproc.residual
        residual_blocks = []
        for i, o, b in zip(
            residual.features.in_features, residual.features.out_features,
            residual.features.bottleneck_features, 
        ):
            residual_blocks.append(mires.make_residual_block(
                block_type=residual.block,
                convolution_type=residual.convolution,
                out_features=o,
                in_features=i,
                bottleneck_features=b,
                activation_type=residual.activation,
                strided=False,
            ))
        self.pre = torch.nn.Sequential(
            *preproc_convs, *residual_blocks,
        )
        branches_config = configuration.branches
        start_trans_config = modules['start_transition']
        self.start_trans = hyu.instantiate(start_trans_config, 
            in_features=residual.features.out_features[-1],
            start_features=branches_config.start_features
        )
        #NOTE: stages
        highres_module = modules['highres'] # NOTE: outputs list of # branches outputs
        self.stages = torch.nn.ModuleList([
            torch.nn.Sequential(*[
                hyu.instantiate(highres_module, 
                    branches=i, depth=d, start_features=branches_config.start_features
                ) for _, d in zip(range(modules), depths)
            ]) for i, modules, depths in zip(
                range(2, configuration.stages + 1),
                branches_config.modules,
                branches_config.depths,
            )
        ])
        stage_trans_config = modules['stage_transition']
        self.stage_transitions = torch.nn.ModuleList([
            hyu.instantiate(stage_trans_config, branches=i + 1,
                prev_branch_features=branches_config.start_features * (2 ** i),
            ) for i in range(1, configuration.stages - 1)
        ])
        head_module = modules['head']
        self.head = hyu.instantiate(head_module,
            stages=configuration.stages,
            start_features=branches_config.start_features,
            out_features=configuration.out_features,
        )
        self.input = ensure_string_list(configuration.input)
        self.output = ensure_string_list(configuration.output)
        self.output_prefix = configuration.output

    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:        
        for i, o in zip(self.input, self.output):
            in_tensor = td[i]
            preproc = self.pre(in_tensor)
            hr_inputs = self.start_trans(preproc)
            for stage, trans in zip(self.stages, self.stage_transitions):
                hr_inputs = trans(stage(hr_inputs))
            prediction = self.head(self.stages[-1](hr_inputs))
            if type(prediction) == list: #NOTE: to support higherhnet                
                for i , heatmap in enumerate(prediction):
                    td[f"{self.output_prefix[:-2]}_{i+1}"] = heatmap
            else:
                td[o] = prediction
        return td


import torch
import functools

__all__ = [
    "SixD2",
    "SixD5",
    "SixD10",
]

def _calculate_add(
    pts_gt:     torch.Tensor,
    pts_est:    torch.Tensor,
    threshold:  float,
    diagonal:   float
):
    """
        Average Distance of Model Points for objects with no indistinguishable views
        - by Hinterstoisser et al. (ACCV 2012).
        :pts_est: transformed pts by employing the predicted pose
        :pts_gt: transformed pts by employing the gt pose
        :return: Error of pose_est w.r.t. pose_gt.
        Then  a  pose estimation is considered to be correct if the computed average distance is within threshold% of the model diagonal.
    """
    b, _, __ = pts_gt.size()  
    #TODO: check the xyz position if is channel or row ?
    e = torch.linalg.norm(pts_est - pts_gt, ord=2, dim=1).mean(dim=-1)
    threshold /= 100
    valid_invalid = torch.where( e < (threshold * diagonal), 1.0 , 0.0)
    valid = torch.count_nonzero(valid_invalid)
    acc = 100.0 * float(valid) / b    
    return torch.tensor(acc).float()

class SixD(torch.nn.Module):
    def __init__(self,
        diagonal:   float,
        threshold:  int = 10,
    ):
        super(SixD, self).__init__()        
        self.threshold = threshold
        self.diagonal = diagonal
    
    def forward(
        self,
        pts_gt: torch.Tensor,
        pts_est: torch.Tensor,
    ) -> torch.Tensor:
        return _calculate_add(
            pts_gt,
            pts_est,
            self.threshold,
            self.diagonal
        )

SixD2 = functools.partial(
    SixD,
    threshold=2
)

SixD5 = functools.partial(
    SixD, 
    threshold=5
)

SixD10 = functools.partial(
    SixD, 
    threshold=10
)


import copy
import torch
import typing

__all__ = ['AccSGD']

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/accsgd.py

class AccSGD(torch.optim.Optimizer):
    r"""Implements AccSGD algorithm.

    - **Paper**: [On the insufficiency of existing momentum schemes for Stochastic Optimization](https://arxiv.org/pdf/1803.05591.pdf)
    - **Paper**: [Accelerating Stochastic Gradient Descent For Least Squares Regression](https://arxiv.org/pdf/1704.08227.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: learning rate (default: 1e-3)
        kappa: ratio of long to short step (default: 1000)
        xi: statistical advantage parameter (default: 10)
        small_const: any value <=1 (default: 0.7)
        weight_decay: weight decay (L2 penalty) (default: 0)

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.AccSGD(model.parameters(), lr=0.1)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

     __ https://arxiv.org/abs/1704.08227
     __ https://arxiv.org/abs/1803.05591

    Note:
        Reference code: https://github.com/rahulkidambi/AccSGD
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        kappa: float=1000.0,
        xi: float=10.0,
        small_const: float=0.7,
        weight_decay: float=0,
    ) -> None:
        if lr <= 0.0:
            raise ValueError('Invalid learning rate: {}'.format(lr))
        if weight_decay < 0:
            raise ValueError(
                'Invalid weight_decay value: {}'.format(weight_decay)
            )
        defaults = dict(
            lr=lr,
            kappa=kappa,
            xi=xi,
            small_const=small_const,
            weight_decay=weight_decay,
        )
        super(AccSGD, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            large_lr = (group['lr'] * group['kappa']) / (group['small_const'])
            alpha = 1.0 - (
                (group['small_const'] * group['small_const'] * group['xi'])
                / group['kappa']
            )
            beta = 1.0 - alpha
            zeta = group['small_const'] / (group['small_const'] + beta)
            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = p.grad.data
                if weight_decay != 0:
                    d_p.add_(p.data, alpha=weight_decay)
                param_state = self.state[p]
                if 'momentum_buffer' not in param_state:
                    param_state['momentum_buffer'] = copy.deepcopy(p.data)
                buf = param_state['momentum_buffer']
                buf.mul_((1.0 / beta) - 1.0)
                buf.add_(d_p, alpha=-large_lr)
                buf.add_(p.data)
                buf.mul_(beta)

                p.data.add_(d_p, alpha=-group['lr'])
                p.data.mul_(zeta)
                p.data.add_(buf, alpha=1.0 - zeta)

        return loss


import moai.networks.lightning as minet
import moai.utils.parsing.rtp as mirtp

from collections import OrderedDict

import torch
import inspect
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging 
import toolz

__all__ = ['MultiBranch']

log = logging.getLogger(__name__)

def _create_processing_block(
    cfg: omegaconf.DictConfig, 
    attribute: str, 
    monads: omegaconf.DictConfig
):
    if not cfg and attribute in cfg:
        log.warning(f"Empty processing block ({attribute}) in feedforward model.")
    return hyu.instantiate(getattr(cfg, attribute), monads)\
        if cfg and attribute in cfg else torch.nn.Identity()

class MultiBranch(minet.FeedForward):
    def __init__(self,
        configuration:  omegaconf.DictConfig, 
        modules:        omegaconf.DictConfig,        
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,
        monads:         omegaconf.DictConfig=None,
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(MultiBranch, self).__init__(
            data=data, parameters=parameters,
            feedforward=feedforward, monads=monads,
            supervision=supervision, validation=validation,
            visualization=visualization, export=export,
        )
        self.encoder = hyu.instantiate(modules.encoder)
        self.decoders = torch.nn.ModuleDict(OrderedDict(
            (k, hyu.instantiate(d)) for k, d in sorted(
                toolz.dissoc(modules, 'encoder').items(), 
                key=lambda kvp: kvp[0]
            )
        ))
        self.enc_fwds = []

        params = inspect.signature(self.encoder.forward).parameters
        enc_in = list(zip(*[mirtp.force_list(configuration.encoder[prop]) for prop in params]))
        enc_out = mirtp.split_as(mirtp.resolve_io_config(configuration.encoder['out']), enc_in)
        
        self.enc_res_fill = [mirtp.get_result_fillers(self.encoder, out) for out in enc_out]        
        get_enc_filler = iter(self.enc_res_fill)
        
        for keys in enc_in:
            self.enc_fwds.append(lambda td,
                tk=keys,
                args=params.keys(),
                enc=self.encoder,
                filler=next(get_enc_filler):
                    filler(td, enc(**dict(zip(args,
                        list(
                                td[k] if type(k) is str
                                else list(td[j] for j in k)
                            for k in tk
                        )
                    ))))
            )

        self.dec_fwds = OrderedDict()
        for name, decoder in self.decoders.items():
            params = inspect.signature(decoder.forward).parameters
            dec_in = list(zip(*[mirtp.force_list(configuration[name][prop]) for prop in params if configuration[name][prop] is not None]))
            #NOTE: dec_in receives them ordered (list), and so args cannot be ignored (nulled) when not last in the list
            #TODO: fix it via mapping
            dec_out = mirtp.split_as(mirtp.resolve_io_config(configuration[name]['out']), dec_in)

            dec_res_fill = [mirtp.get_result_fillers(decoder, out) for out in dec_out]
            get_dec_filler = iter(dec_res_fill)
            
            self.dec_fwds[name] = []
            for keys in dec_in:
                self.dec_fwds[name].append(lambda td,
                    tk=keys,
                    args=params.keys(), 
                    dec=decoder,
                    filler=next(get_dec_filler):
                        filler(td, dec(**dict(zip(args,
                            list(
                                    td[k] if type(k) is str
                                    else list(td[j] for j in k)
                                for k in tk
                            )
                        ))))
                )
        self.pre_decode = torch.nn.ModuleList(
            _create_processing_block(feedforward, f'pre_decode{i}', monads)\
                for i in range(1, len(self.decoders) + 1)
        )

    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        for e in self.enc_fwds:
            e(td)
        for ((k, v), pre) in zip(
            self.dec_fwds.items(),
            self.pre_decode,
        ):
            for d in v:
                td = pre(td)
                d(td)
        return td


from moai.visualization.visdom.base import Base
from moai.utils.color.colorize import get_colormap, COLORMAPS
from moai.utils.arguments import assert_numeric
from torchvision.utils import make_grid

import torch
import visdom
import functools
import typing
import logging
import numpy

log = logging.getLogger(__name__)

__all__ = ["Image_grid2d"]

class Image_grid2d(Base):
    def __init__(self,
        keys:           typing.Union[str, typing.Sequence[str]],
        types:          typing.Union[str, typing.Sequence[str]],
        colormaps:      typing.Union[str, typing.Sequence[str]],
        transforms:     typing.Union[str, typing.Sequence[str]],
        name:           str="default",
        ip:             str="http://localhost",
        port:           int=8097,   
    ):
        super(Image_grid2d, self).__init__(name, ip, port)
        self.keys = [keys] if type(keys) is str else list(keys)
        self.types = [types] if type(types) is str else list(types)
        self.transforms = [transforms] if type(transforms) is str else list(transforms)
        self.colormaps = [colormaps] if type(colormaps) is str else list(colormaps)
        self.viz_map = {
            'color_grid': functools.partial(self.__viz_color, self.visualizer)
        }
        self.transform_map = {
            'none': functools.partial(self.__no_transform),
            'minmax': functools.partial(self.__minmax_normalization),
            'ndc': functools.partial(self.__ndc_to_one),
        }
        self.colorize_map = { "none": lambda x: x }
        self.colorize_map.update(COLORMAPS)

    @property
    def name(self) -> str:
        return self.env_name

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        _, _, ch, w, h = tensors.shape
        for t, tf, c in zip(self.types, self.transforms, self.colormaps):
            self.viz_map[t](
                self.colorize_map[c](
                    self.transform_map[tf](
                        make_grid(tensors.reshape(-1, ch, w, h), nrow=7)
                    )
                ), 'color_grid', 'color_grid', self.name
            )

    @staticmethod
    def __viz_color(
        visdom: visdom.Visdom,
        tensor: torch.Tensor,
        key: str,
        win: str,
        env: str
    ) -> None:
        visdom.images(
            tensor,
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': 50,
            }
        )

    @staticmethod
    def __no_transform(tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    @staticmethod
    def __minmax_normalization(tensor: torch.Tensor) -> torch.Tensor:
        _, __, ___ = tensor.size()
        min_v = torch.min(tensor)
        max_v = torch.max(tensor)
        return (tensor - min_v) / (max_v - min_v)

    @staticmethod
    def __ndc_to_one(tensor: torch.Tensor) -> torch.Tensor:
        return torch.addcmul(
            torch.scalar_tensor(0.5).to(tensor), 
            tensor,
            torch.scalar_tensor(0.5).to(tensor)
        )


import torch
import numpy as np

__all__ = ["CosineDistance"]

#NOTE: see https://github.com/pytorch/pytorch/issues/8069
#TODO: update acos_safe once PR mentioned in above link is merged and available

def _acos_safe(x: torch.Tensor, eps: float=1e-4):
    slope = np.arccos(1.0 - eps) / eps
    # TODO: stop doing this allocation once sparse gradients with NaNs (like in
    # th.where) are handled differently.
    buf = torch.empty_like(x)
    good = torch.abs(x) <= 1.0 - eps
    bad = ~good
    sign = torch.sign(x[bad])
    buf[good] = torch.acos(x[good])
    buf[bad] = torch.acos(sign * (1.0 - eps)) - slope * sign * (torch.abs(x[bad]) - 1.0 + eps)
    return buf

'''
    def _acos_safe(x: torch.Tensor, eps: float=1e-4):
        sign = torch.sign(x)
        slope = np.arccos(1.0 - eps) / eps
        return torch.where(torch.abs(x) <= 1.0 - eps,
            torch.acos(x),
            torch.acos(sign * (1.0 - eps)) - slope * sign * (torch.abs(x) - 1.0 + eps)
        )
'''

class CosineDistance(torch.nn.CosineSimilarity):
    def __init__(self,
        dim:                    int=1,
        epsilon:                float=1e-4,
        normalized:             bool=True,
    ):
        super(CosineDistance, self).__init__(dim=dim, eps=epsilon)
        self.normalized = normalized
        self.epsilon = epsilon

    def forward(self,
        pred: torch.Tensor,
        gt: torch.Tensor,        
        weights: torch.Tensor=None, # float tensor
        mask: torch.Tensor=None, # byte tensor
    ) -> torch.Tensor:        
        dot = torch.sum(gt * pred, dim=self.dim) if self.normalized\
            else super(CosineDistance, self).forward(pred, gt)
        # return torch.acos(dot) / np.pi #NOTE: (eps) clamping should also fix the nan grad issue (traditional [-1, 1] clamping does not)
        # return torch.acos(torch.clamp(dot, min=-1.0 + self.epsilon, max=1.0 - self.epsilon)) / np.pi
        return _acos_safe(dot, eps=self.epsilon) / np.pi        



import torch
import typing
import sys

#NOTE: assumes ldr images normalized in a specific range: set 1.0 for values in [0,255]
class PSNR(torch.nn.Module):
    def __init__(self,
        range_scale_factor: float=255.0
    ):
        super(PSNR, self).__init__()
        self.dr = range_scale_factor            
        self.eps = torch.scalar_tensor(sys.float_info.epsilon) # https://stackoverflow.com/questions/9528421/value-for-epsilon-in-python 

    def forward(self,
        gt: torch.Tensor,
        pred: torch.Tensor,
    ) -> torch.Tensor:
        mse = torch.mean(torch.sub(self.dr * gt, self.dr * pred) ** 2)
        return 10 * torch.log10((self.dr * self.dr) / (mse + self.eps.to(mse)))


# Copyright 2020-present, ai in motion
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import os
import re
import io
import logging

from setuptools import (
    setup,
    find_packages
)
from pkg_resources import (
    DistributionNotFound,
    get_distribution, 
    parse_version,
)

logger = logging.getLogger()
logging.basicConfig(format='%(levelname)s - %(message)s')

def get_readme():
    base_dir = os.path.abspath(os.path.dirname(__file__))
    with io.open(os.path.join(base_dir, "README.md"), encoding="utf-8") as f:
        return f.read()

def get_requirements():
    base_dir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(base_dir, "requirements.txt"), encoding="utf-8") as f:
        return [line.strip() for line in f.readlines()]

def moai_info():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    version_file = os.path.join(current_dir, "moai", "__init__.py")
    with io.open(version_file, encoding="utf-8") as f:
        version = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)
        f.seek(0)
        author = re.search(r'^__author__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)
        f.seek(0)
        email = re.search(r'^__author_email__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)
        f.seek(0)
        licence = re.search(r'^__license__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)
        f.seek(0)
        url = re.search(r'^__homepage__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)
        f.seek(0)
        code_url = re.search(r'^__github_repo__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)
        f.seek(0)
        docs_url = re.search(r'^__documentation_page__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)
        f.seek(0)
        short_desc = re.search(r'^__docs__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)
        f.seek(0)
        keywords = re.search(r'^__keywords__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)
        return version, author, email, licence, \
            url, code_url, docs_url, \
            short_desc, keywords.split(',')

PACKAGE_NAME = 'moai-mdk'
VERSION, AUTHOR, EMAIL, LICENSE,\
URL, CODE_URL, DOCS_URL, \
DESCRIPTION, KEYWORDS = moai_info()

if __name__ == '__main__':
    logger.info(f"Installing {PACKAGE_NAME} (v: {VERSION}) ...")
    setup(
        name=PACKAGE_NAME,
        version=VERSION,
        author=AUTHOR,
        author_email=EMAIL,
        description=DESCRIPTION,
        long_description=get_readme(),
        long_description_content_type="text/markdown",
        keywords=KEYWORDS,
        licence_file='LICENCE',
        url=URL,
        project_urls={
            'Documentation': DOCS_URL,
            'Source': CODE_URL,
        },
        packages=find_packages(exclude=('docs', 'outputs')),
        install_requires=get_requirements(),
        include_package_data=True,
        python_requires='~=3.7',
        package_dir={'moai': 'moai'},
        package_data={'moai': ['conf/**/*.yaml']},
        entry_points={
            'console_scripts': [
                'moai=moai.__main__:moai',
                # 'moai=moai.__main__',
            ],
        },
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 3.7",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Software Development :: Libraries",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],    
    )


from moai.export.local.image2d import Image2d
from moai.monads.execution.cascade import _create_accessor

import torch
import pyrender
import typing
import logging
import numpy as np
import math
import trimesh
import itertools
from PIL import Image

log = logging.getLogger(__name__)

__all__ = ["RenderedMesh"]

class RenderedMesh(Image2d):
    def __init__(self,
        path:               str,
        vertices:           typing.Union[str, typing.Sequence[str]],
        faces:              typing.Union[str, typing.Sequence[str]],
        image:              typing.Union[str, typing.Sequence[str]],
        colormap:           typing.Union[str, typing.Sequence[str]],
        transform:          typing.Union[str, typing.Sequence[str]],
        translation:        typing.Union[str, typing.Sequence[str]]=None,
        rotation:           typing.Union[str, typing.Sequence[str]]=None,
        focal_length:       typing.Union[float, typing.Tuple[float, float]]=5000.0,
        extension:  typing.Union[str, typing.Sequence[str]]=["png"], # jpg or png or exr
        scale:              float=1.0,
        batch_percentage:   float=1.0,        
    ):
        super(RenderedMesh, self).__init__(
            path=path, image=image, extension=extension,
            type=list(itertools.repeat('color', len([vertices] if isinstance(vertices, str) else vertices))),
            transform=transform, batch_percentage=batch_percentage, 
            colormap=colormap,
        )
        self.focal_length = (float(focal_length), float(focal_length)) \
            if isinstance(focal_length, float) or isinstance(focal_length, int) else focal_length
        self.material = pyrender.MetallicRoughnessMaterial(
            metallicFactor=0.0, alphaMode='OPAQUE', baseColorFactor=(1.0, 1.0, 0.9, 1.0)
        )
        self.vertices = [vertices] if isinstance(vertices, str) else list(vertices)
        self.vertices = [_create_accessor(k) for k in self.vertices]
        self.faces = [faces] if isinstance(faces, str) else list(faces)
        self.faces = [_create_accessor(k) for k in self.faces]
        self.scene = pyrender.Scene(
            bg_color=[0.0, 0.0, 0.0, 0.0],
            ambient_light=(0.3, 0.3, 0.3)
        )
        for light in self._create_raymond_lights():
            self.scene.add_node(light)
        self.translation = list(itertools.repeat('', len(self.keys)) if translation is None else\
            ([translation] if isinstance(translation, str) else list(translation)))
        self.rotation = list(itertools.repeat('', len(self.keys)) if rotation is None else\
            ([rotation] if isinstance(rotation, str) else list(rotation)))
        self.scale = scale
        self.renderer = None

    def _get_renderer(self, width: int, height: int) -> pyrender.OffscreenRenderer:
        if self.renderer is None or self.renderer.viewport_width != width\
            or self.renderer.viewport_height != height:
                self.renderer = pyrender.OffscreenRenderer(
                    viewport_width=width, viewport_height=height, point_size=1.0
                )                    
        return self.renderer

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for v, f, r, t, k, _, tf, c, fmt in zip(
            self.vertices, self.faces, self.rotation, self.translation,
            self.keys, self.types, self.transforms, self.colormaps, self.formats
        ):
            take = int(math.ceil(self.batch_percentage * tensors[k].shape[0]))
            background = self.colorize_map[c](
                self.transform_map[tf](tensors, k, take)
            )
            b, c, h, w = background.shape
            renderer = self._get_renderer(width=w, height=h)
            results = []
            for i in range(b):
                rotation = tensors[r][i].detach().cpu().numpy().squeeze() if r else np.eye(3)
                translation = tensors[t][i].detach().cpu().numpy().squeeze() if t else np.zeros(3)

                tmesh = trimesh.Trimesh(
                    v(tensors).detach().cpu().numpy().squeeze(),
                    f(tensors).detach().cpu().numpy().squeeze(),
                    process=False
                )
                rot = trimesh.transformations.rotation_matrix(np.radians(180), [1, 0, 0])
                tmesh.apply_transform(rot)
                mesh = pyrender.Mesh.from_trimesh(tmesh, material=self.material)
                node = self.scene.add(mesh, 'mesh')

                # Equivalent to 180 degrees around the y-axis. Transforms the fit to
                # OpenGL compatible coordinate system.
                translation[0] *= -1.0
                camera_pose = np.eye(4)
                camera_pose[:3, :3] = rotation
                camera_pose[:3, 3] = translation

                camera = pyrender.camera.IntrinsicsCamera(
                    fx=self.focal_length[0], cx=w // 2,
                    fy=self.focal_length[1], cy=h // 2,
                )
                cam = self.scene.add(camera, pose=camera_pose)

                color, _ = renderer.render(self.scene, flags=pyrender.RenderFlags.RGBA)
                color = color.astype(np.float32) / 255.0

                valid_mask = (color[:, :, -1] > 0)[:, :, np.newaxis]
                input_img = background.detach().cpu().numpy().squeeze().transpose(1, 2, 0)
                output_img = (color[:, :, :-1] * valid_mask + (1 - valid_mask) * input_img)
                if self.scale != 1.0:
                    output_img = np.array(
                        Image.fromarray(
                            (output_img * 255.0).astype(np.uint8)
                        ).resize(
                            (int(w * self.scale), int(h * self.scale)), Image.ANTIALIAS
                        )
                    )
                results.append(output_img)
                self.scene.remove_node(node)
                self.scene.remove_node(cam)
            self.save_map['color'](
                np.stack(results).transpose(0, 3, 1, 2),
                f"{k}_overlay", self.index, fmt
            )
        self.index = 0 if self.mode == "overwrite" else self.index + b


    def _create_raymond_lights(self):
        thetas = np.pi * np.array([1.0 / 6.0, 1.0 / 6.0, 1.0 / 6.0])
        phis = np.pi * np.array([0.0, 2.0 / 3.0, 4.0 / 3.0])

        nodes = []

        for phi, theta in zip(phis, thetas):
            xp = np.sin(theta) * np.cos(phi)
            yp = np.sin(theta) * np.sin(phi)
            zp = np.cos(theta)

            z = np.array([xp, yp, zp])
            z = z / np.linalg.norm(z)
            x = np.array([-z[1], z[0], 0.0])
            if np.linalg.norm(x) == 0:
                x = np.array([1.0, 0.0, 0.0])
            x = x / np.linalg.norm(x)
            y = np.cross(z, x)

            matrix = np.eye(4)
            matrix[:3,:3] = np.c_[x,y,z]
            nodes.append(pyrender.Node(
                light=pyrender.DirectionalLight(color=np.ones(3), intensity=1.0),
                matrix=matrix
            ))

        return nodes



from moai.monads.generation.tensor.torch import (
    Scalar,
    Random,
    Ones,
    Zeros,
    Clone,
    Parameter,
    Parameters,
)

__all__ = [
    "Scalar",
    "Random",
    "Ones",
    "Zeros",
    "Clone",
    "Parameter",
    "Parameters",
]


from moai.visualization.visdom.image2d import Image2d
from moai.visualization.visdom.blend2d import Blend2d
from moai.visualization.visdom.feature2d import Feature2d
from moai.visualization.visdom.gizmo2d import Gizmo2d
from moai.visualization.visdom.pose2d import Pose2d
from moai.visualization.visdom.image_grid2d import Image_grid2d
# from moai.visualization.visdom.rendered_mesh import RenderedMesh
from moai.visualization.visdom.vector import Vector
    
__all__ = [
    "Image2d",
    "Blend2d",
    "Feature2d",
    "Gizmo2d",
    "Pose2d",
    "Image_grid2d",
    # "RenderedMesh",
    "Vector",
]


import torch

from moai.monads.utils import (
    spatial_dims
)

import functools

__all__ = [
    "QuantizeCoords",
]

class QuantizeCoords(torch.nn.Module):
    def _convert_norm(self, 
        uvs:                torch.Tensor, 
        height:             int,
        width:              int,
    ) -> torch.Tensor:
        dims = torch.Tensor([width, height]).to(uvs)
        return uvs / dims

    def _convert_ndc(self, 
        uvs:                torch.Tensor, 
        height:             int,
        width:              int,
    ) -> torch.Tensor:
        one = self._convert_one(uvs, height, width)
        return one * 2.0 - 1.0

    def __init__(self,        
        mode:       str='round', # round, ceil, floor
        flip:       bool= False,
        coord_type: str='ndc',
        width:      int=1,
        height:     int=1,
    ):
        super(QuantizeCoords,self).__init__()
        self.mode = mode
        self.flip = flip
        self.coord_type = coord_type
        self.width = width
        self.height = height
        if self.mode == 'round':
            self.branch = functools.partial(torch.round)
        elif self.mode == 'ceil':
            self.branch = functools.partial(torch.ceil)
        elif self.mode == 'floor':
            self.branch = functools.partial(torch.floor)
        else:
            raise ValueError("{} mode is not valid.".format(self.mode))

    def forward(self,
        coords:     torch.Tensor,
        grid:       torch.Tensor=None,
    ) -> torch.Tensor:
        if self.coord_type != 'coord':
            coords[..., 0] = (coords[..., 0] + 1.0) / 2.0 * self.width if self.coord_type == 'ndc' else coords[..., 0] * self.width
            coords[..., 1] = (coords[..., 1] + 1.0) / 2.0 * self.height if self.coord_type == 'ndc' else coords[..., 1] * self.height
        coords = self.branch(coords)

        if self.coord_type != 'coord':
            coords = self._convert_ndc(coords, self.height, self.width) if self.coord_type == 'ndc' \
                else self._convert_norm(coords, self.height, self.width)
        # if grid is not None:
        #     dims = spatial_dims(grid)
        #     for b in range(coords.shape[0]):                
        #         for j in range(coords.shape[1]):
        #             for d in range(len(dims)):
        #                 coords[b, j, d] = grid[b, d][tuple(coords.int()[b, j].flip(-1))]
        return coords.flip(-1) if self.flip else coords


from moai.utils.arguments import assert_numeric
from moai.monads.execution.cascade import _create_accessor
from collections.abc import Callable

import scenepic
import httpwatcher
import typing
import torch
import logging
import colour
import toolz
import math
import os
import numpy as np

log = logging.getLogger(__name__)

__all__ = ['Mesh']

class Mesh(Callable):    
    def __init__(self,
        vertices:         typing.Union[str, typing.Sequence[str]],
        faces:            typing.Union[str, typing.Sequence[str]],
        canvas:           typing.Union[int, typing.Sequence[int]],
        layer:            typing.Union[int, typing.Sequence[int]],
        color:            typing.Union[str, typing.Sequence[str]],
        batch_percentage: float=1.0,
        width:            int=600,
        height:           int=400,
        name:             str="default",        
    ):
        self.name= name
        self.vertices = [vertices] if isinstance(vertices, str) else list(vertices)
        self.faces = [faces] if isinstance(faces, str) else list(faces)
        self.vertex_accessors = [_create_accessor(k) for k in self.vertices]        
        self.face_accesors = [_create_accessor(k) for k in self.faces]
        self.ids = [canvas] if isinstance(canvas, int) else list(canvas)
        self.layers = [layer] if isinstance(layer, int) else list(layer)
        self.colors = [colour.Color(color)] if isinstance(color, str) else list(colour.Color(c) for c in color)
        self.batch_percentage = batch_percentage
        assert_numeric(log, 'batch percentage', batch_percentage, 0.0, 1.0)        
        self.WH = (width, height)
        scenepic_folder = os.path.join(os.getcwd(), 'scenepic')
        os.makedirs(scenepic_folder, exist_ok=True)
        log.info(f"Scenepic visualization enabled @ {scenepic_folder}.")
        log.warning(f"[scenepic]: For automatic refreshing @ `http://localhost:5555' use `httpwatcher -r {scenepic_folder}` (`pip install httpwatcher` if not available)")

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        meshes = {}
        scene = scenepic.Scene()
        canvases = [
            scene.create_canvas_3d(width=self.WH[0], height=self.WH[1]) 
            for _ in toolz.unique(self.ids)
        ]
        
        for n, v, f, c, l, id in zip(self.vertices, self.vertex_accessors,
            self.face_accesors, self.colors, self.layers, self.ids
        ):
            vertices = v(tensors).detach().cpu().numpy()
            faces = f(tensors).detach().cpu().numpy()
            b = math.ceil(self.batch_percentage * vertices.shape[0])
            if id not in meshes:
                meshes[id] = []
            for i in range(b):
                mesh = scene.create_mesh(mesh_id=f"{n}_{i}",
                    shared_color=scenepic.Color(*c.rgb), layer_id=f"{l}",
                )
                mesh.add_mesh_without_normals(vertices[i], faces[i])
                meshes[id].append(mesh)
        for k, m in meshes.items():
            frame = canvases[k].create_frame()
            grouped = toolz.groupby(lambda x: x.layer_id, m)
            for l, m in grouped.items():
                for j, mesh in enumerate(m):
                    xform = scenepic.Transforms.Translate(j * np.array([1.0, 0.0, 0.0]))
                    frame.add_mesh(mesh, xform)
            frame.set_layer_settings(dict(
                (str(n), {'opacity': 0.5}) for n in toolz.unique(self.layers)
            ))
        scene.link_canvas_events(*canvases)
        scene.save_as_html(os.path.join('scenepic', "index.html"), title=f"{self.name}")


from moai.monads.utils import spatial_dim_list

import torch

__all__ = ["CenterOfMass"]

class CenterOfMass(torch.nn.Module):
    def __init__(self,
        mode: str="default", #TODO: to support MoM, DCoM
        flip: bool=True, # flip order of coordinates as originated from the grid's order
    ):
        super(CenterOfMass, self).__init__()
        self.mode = mode
        self.flip = flip

    def _extract_center_of_mass(self,
        grid: torch.Tensor,     # [B, (S)VU, (D), H, W], order is (S)VU not UV(S), y coord first channel
        heatmaps: torch.Tensor  # [B, K, (D), H, W], with its value across the spatial dimensions summing to unity
    ) -> torch.Tensor:          # [B, K, UV(S) or (S)VU]
        channels = heatmaps.shape[1]        
        sum_dims = spatial_dim_list(heatmaps)
        CoMs = []
        for i in range(channels): #TODO: refactor for loop
            heatmap = heatmaps[:, i, ...].unsqueeze(1)
            CoM = torch.sum(heatmap * grid, dim=sum_dims)
            CoMs.append(CoM)
        return torch.stack(CoMs, dim=1)
                
    def forward(self, 
        grid: torch.Tensor,     # coordinates grid tensor of C coordinates
        heatmaps: torch.Tensor, # spatial probability tensor of K keypoints
    ) -> torch.Tensor:
            coms = self._extract_center_of_mass(grid, heatmaps)
            return torch.flip(coms, dims=[2]) if self.flip else coms


import moai.utils.engine as mieng

import omegaconf.omegaconf
import logging

log = logging.getLogger(__name__)

__all__ = ["Exporter"]

class Exporter(mieng.Interval, mieng.Single):
    def __init__(self,
        batch_interval:         int,
        exporters:              omegaconf.DictConfig,        
    ):
        mieng.Interval.__init__(self, batch_interval)
        mieng.Single.__init__(self, items=exporters, name="exporter")
        log.info(f"Exporting data at a {batch_interval} step interval.")


import albumentations
import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["OneOf"]

class OneOf(torch.utils.data.Dataset):
    def __init__(self,
        dataset:        torch.utils.data.Dataset,
        inputs:         typing.Sequence[str],
        outputs:        typing.Sequence[str],
        augmentations:  omegaconf.DictConfig={},
        probability:    float=0.5,
    ):
        super(OneOf, self).__init__()
        self.inner, self.inputs, self.outputs = dataset, inputs, outputs
        self.augmentations = []
        for k, v in augmentations.items():
            log.info(f"Using {k} augmentation /w a {v.p * 100.0}% probability.")
            self.augmentations.append(hyu.instantiate(v))
        self.oneof = albumentations.OneOf(self.augmentations, p=probability)
        weight_sum = sum(list(map(lambda a: a.p, self.augmentations)), 0.0)
        if weight_sum == 0.0:
            log.warning("OneOf weights are zero, reverting to a NoOp.")
            self.oneof = albumentations.OneOf([albumentations.NoOp()], p=1.0)
        elif weight_sum != 1.0:
            log.warning(f"OneOf augmentation weights do not sum up to unity ({weight_sum}), they will be normalized to unity.")

    def __len__(self) -> int:
        return len(self.inner)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        item = self.inner[index]
        for i, o in zip(self.inputs, self.outputs):
            data = {"image": item[i].numpy().transpose(1, 2, 0)} # numpy conversion overhead is insignificant
            augmented = self.oneof(**data)
            item[o] = torch.from_numpy(augmented["image"].transpose(2, 0, 1))
        return item


import torch
import typing

__all__ = ['QHAdam']

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/qhadam.py

class QHAdam(torch.optim.Optimizer):
    r"""Implements the QHAdam optimization algorithm.

    - **Paper**: [QUASI-HYPERBOLIC MOMENTUM AND ADAM FOR DEEP LEARNING](https://arxiv.org/pdf/1810.06801.pdf)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: learning rate (default: 1e-3)
        betas: coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        nus: immediate discount factors used to estimate the gradient and its
            square (default: (1.0, 1.0))
        eps: term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay: weight decay (L2 penalty) (default: 0)
        decouple_weight_decay: whether to decouple the weight
            decay from the gradient-based optimization step (default: False)

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.QHAdam(model.parameters(), lr=0.1)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ https://arxiv.org/abs/1810.06801

    Note:
        Reference code: https://github.com/facebookresearch/qhoptim
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float = 1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        nus: typing.Tuple[float, float]=(1.0, 1.0),
        weight_decay: float=0.0,
        decouple_weight_decay: bool=False,
        eps: float=1e-8,
    ):
        if lr <= 0.0:
            raise ValueError('Invalid learning rate: {}'.format(lr))
        if eps < 0.0:
            raise ValueError('Invalid epsilon value: {}'.format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(
                'Invalid beta parameter at index 0: {}'.format(betas[0])
            )
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(
                'Invalid beta parameter at index 1: {}'.format(betas[1])
            )
        if weight_decay < 0:
            raise ValueError(
                'Invalid weight_decay value: {}'.format(weight_decay)
            )

        defaults = {
            'lr': lr,
            'betas': betas,
            'nus': nus,
            'weight_decay': weight_decay,
            'decouple_weight_decay': decouple_weight_decay,
            'eps': eps,
        }
        super(QHAdam, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group['lr']
            beta1, beta2 = group['betas']
            nu1, nu2 = group['nus']
            weight_decay = group['weight_decay']
            decouple_weight_decay = group['decouple_weight_decay']
            eps = group['eps']

            for p in group['params']:
                if p.grad is None:
                    continue

                d_p = p.grad.data
                if d_p.is_sparse:
                    raise RuntimeError(
                        'QHAdam does not support sparse gradients, '
                        'please consider SparseAdam instead'
                    )

                state = self.state[p]

                if weight_decay != 0:
                    if decouple_weight_decay:
                        p.data.mul_(1 - lr * weight_decay)
                    else:
                        d_p.add_(p.data, alpha=weight_decay)

                d_p_sq = d_p.mul(d_p)

                if len(state) == 0:
                    state['beta1_weight'] = 0.0
                    state['beta2_weight'] = 0.0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)

                state['beta1_weight'] = 1.0 + beta1 * state['beta1_weight']
                state['beta2_weight'] = 1.0 + beta2 * state['beta2_weight']

                beta1_weight = state['beta1_weight']
                beta2_weight = state['beta2_weight']
                exp_avg = state['exp_avg']
                exp_avg_sq = state['exp_avg_sq']

                beta1_adj = 1.0 - (1.0 / beta1_weight)
                beta2_adj = 1.0 - (1.0 / beta2_weight)
                exp_avg.mul_(beta1_adj).add_(d_p, alpha=1.0 - beta1_adj)
                exp_avg_sq.mul_(beta2_adj).add_(d_p_sq, alpha=1.0 - beta2_adj)

                avg_grad = exp_avg.mul(nu1)
                if nu1 != 1.0:
                    avg_grad.add_(d_p, alpha=1.0 - nu1)

                avg_grad_rms = exp_avg_sq.mul(nu2)
                if nu2 != 1.0:
                    avg_grad_rms.add_(d_p_sq, alpha=1.0 - nu2)
                avg_grad_rms.sqrt_()
                if eps != 0.0:
                    avg_grad_rms.add_(eps)

                p.data.addcdiv_(avg_grad, avg_grad_rms, value=-lr)

        return loss


from moai.utils.arguments.common import (
    assert_numeric,
    assert_non_negative,
    assert_negative,
    assert_positive,
)
from moai.utils.arguments.choices import (
    assert_choices,
    ensure_choices,
)
from moai.utils.arguments.list import (
    ensure_numeric_list,
    ensure_string_list,
    assert_sequence_size,
)
from moai.utils.arguments.path import (
    assert_path,
    ensure_path,
)

__all__ = [
    "assert_numeric",
    "ensure_numeric_list",
    "ensure_string_list",
    "assert_choices",
    "ensure_choices",
    "assert_sequence_size",
    "assert_non_negative",
    "assert_negative",
    "assert_positive",
    "assert_path",
    "ensure_path",
]


from kornia.filters import (
    GaussianBlur2d
)

import torch
import logging
import functools
import typing

log = logging.getLogger(__name__)

__all__ = ["Gaussian2d"]

#NOTE: https://dsp.stackexchange.com/questions/10057/gaussian-blur-standard-deviation-radius-and-kernel-size

class Gaussian2d(GaussianBlur2d):
    def __init__(self,
        kernel_size:                typing.Tuple[int, int]=(5, 5),
        sigma:                      typing.Tuple[float, float]=(1.0, 1.0),
        border_type:                str='reflect'
    ):
        super(Gaussian2d, self).__init__(
            kernel_size=tuple(kernel_size),
            sigma=tuple(sigma), 
            border_type=border_type
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return super(Gaussian2d, self).forward(image)



from moai.monads.utils import (
    expand_spatial_dims,
    spatial_dim_list
)

import torch
import math

__all__ = ["IsotropicGaussian"]

#TODO: for smaller std see https://dsp.stackexchange.com/questions/23460/how-to-calculate-gaussian-kernel-for-a-small-support-size 
class IsotropicGaussian(torch.nn.Module):
    __C__ = math.sqrt(math.pi * 2.0)
    __GRID_TYPE_CHOICES__ = ['ndc', 'coord', 'norm']

    def __init__(self,
        std:            float=1.5, # in % of pixels diagonal, relative to grid => can use varying grid inputs
        normalize:      bool=True,
        scale:          bool=False,
        grid_type:      str='ndc', # 'ndc', 'coord', 'norm'
    ):
        super(IsotropicGaussian, self).__init__()
        self.grid_type = grid_type
        std = std / 100.0 * (2.0 if self.grid_type == 'ndc' else 1.0)
        self.register_buffer("std", torch.scalar_tensor(std))
        self.normalize = normalize
        self.scale = scale

    def forward(self,
        keypoints:  torch.Tensor,       # [B, K, (S)UV or UV(S)] with K the number of keypoints
        grid:       torch.Tensor,       # [B, (S)UV or UV(S), (D), H, W]
        stds:       torch.Tensor=None,  # [B, K]
    ) -> torch.Tensor:                  # [B, K, (D), H, W]
        grid_size = torch.Tensor([*grid.shape[2:]])
        scale = 2.0 if self.grid_type == 'ndc' else 1.0
        std = scale * (stds if stds is not None else self.std)
        if self.grid_type == 'coord':                
            std = std * float(torch.Tensor(grid_size).norm(p=2))
        inv_denom = -0.5 * torch.reciprocal(std ** 2)
        centroids = expand_spatial_dims(keypoints, grid)
        diffs = grid.unsqueeze(1) - centroids
        dist = diffs ** 2
        nom = torch.sum(dist, dim=2)
        gaussian = torch.exp(nom * inv_denom)
        if self.scale: # generate a properly scaled Gaussian
            scaling_factor = torch.reciprocal((IsotropicGaussian.__C__ * std) ** len(grid_size))
            scaling_factor = expand_spatial_dims(scaling_factor, grid)
            gaussian = gaussian * scaling_factor
        if self.normalize: # generate a normalized Gaussian summing to unity
            norm_dims = spatial_dim_list(grid)
            gaussian = gaussian / torch.sum(
                gaussian, dim=norm_dims, keepdim=True
            )
        return gaussian


import torch

__all__ = ['Rotate']

class Rotate(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self,
        rotation:   torch.Tensor,
        points:     torch.Tensor,
    ) -> torch.Tensor:
        return torch.einsum('bij,bpj->bpi', rotation, points)



from moai.visualization.visdom.base import Base
from moai.monads.execution.cascade import _create_accessor

import math
import torch
import typing
import logging
import numpy as np
import visdom
import functools

log = logging.getLogger(__name__)

__all__ = ["Vector"]

class Vector(Base):
    def __init__(self,
        vector:             typing.Union[str, typing.Sequence[str]],
        type:               typing.Union[str, typing.Sequence[str]],
        name:               str="default",
        ip:                 str="http://localhost",
        port:               int=8097,
        batch_percentage:   float=1.0,
    ):
        super(Vector, self).__init__(name=name, ip=ip, port=port)
        self.vector = [vector] if isinstance(vector, str) else list(vector)
        self.vector = [_create_accessor(k) for k in self.vector]
        self.types = [type] if isinstance(type, str) else list(type)
        self.viz_map = {
            'box': functools.partial(self._viz_box, self.visualizer),
        }
        self.batch_percentage = batch_percentage
    
    @property
    def name(self) -> str:
        return self.env_name

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for v, t in zip(
            self.vector, self.types
        ):
            vector = v(tensors)
            b = int(math.ceil(self.batch_percentage * vector.shape[0]))
            for i in range(b):
                self.viz_map[t](
                    vector[i].detach().cpu().numpy(),
                    f"boxplot_{i}", t, self.name
                )
                
    @staticmethod
    def _viz_box(
        visdom: visdom.Visdom,
        array: np.array,
        key: str,
        win: str,
        env: str
    ) -> None:
        visdom.bar(
            array,
            win=f"{key}_{win}",
            env=env,
        )



import torch
import functools

__all__ = ["Taylor"]

class Taylor(torch.nn.Module):
    def __init__(self,
        grid_type:      str='ndc', # 'ndc', 'coord', 'norm'
        flip:           bool=True, # flip order of coordinates as originated from the grid's order
    ):
        super(Taylor, self).__init__()
        self.flip = flip
        self.grid_type = grid_type

    def _convert_norm(self, 
        uvs:                torch.Tensor, 
        height:             int,
        width:              int,
    ) -> torch.Tensor:
        dims = torch.Tensor([width, height]).to(uvs)
        return uvs / dims

    def _convert_ndc(self, 
        uvs:                torch.Tensor, 
        height:             int,
        width:              int,
    ) -> torch.Tensor:
        one = self._convert_one(uvs, height, width)
        return one * 2.0 - 1.0

    def _taylor_coord(self,
            coord: torch.Tensor,        # [UV]
            hm: torch.Tensor,           # [H, W], with its value across the spatial dimensions summing to unity
        ) -> torch.Tensor:              # [UV]
        heatmap_height = hm.shape[0]
        heatmap_width = hm.shape[1]
        px = int(coord[0])
        py = int(coord[1])
        if 1 < px < heatmap_width-2 and 1 < py < heatmap_height-2:
            dx  = 0.5 * (hm[py][px+1] - hm[py][px-1])
            dy  = 0.5 * (hm[py+1][px] - hm[py-1][px])
            dxx = 0.25 * (hm[py][px+2] - 2 * hm[py][px] + hm[py][px-2])
            dxy = 0.25 * (hm[py+1][px+1] - hm[py-1][px+1] - hm[py+1][px-1] \
                + hm[py-1][px-1])
            dyy = 0.25 * (hm[py+2*1][px] - 2 * hm[py][px] + hm[py-2*1][px])

            derivative = torch.tensor([[dx],[dy]]).to(coord)
            hessian = torch.tensor([[dxx,dxy],[dxy,dyy]]).to(coord)
            if dxx * dyy - dxy ** 2 != 0:
                hessianinv = torch.inverse(hessian)
                offset = -hessianinv @ derivative
                offset = torch.squeeze(torch.transpose(offset, 0, 1), dim=0)
                coord += offset

        return coord

    def _taylor(self,
            coords: torch.Tensor,       # [B, K, UV(S) or (S)VU]
            heatmaps: torch.Tensor,     # [B, K, (D), H, W], with its value across the spatial dimensions summing to unity
        ) -> torch.Tensor:              # [B, K, UV(S) or (S)VU]
            channels = heatmaps.shape[1]            
            heatmaps[heatmaps == 0] = 1e-10
            heatmaps = torch.log(heatmaps)                            
            for n in range(coords.shape[0]):
                for p in range(coords.shape[1]):
                    coords[n, p] = self._taylor_coord(coords[n, p], heatmaps[n, p])
            return coords
                
    def forward(self, 
        coords:         torch.Tensor,       # coordinates grid tensor of C coordinates
        heatmaps:       torch.Tensor,       # spatial probability tensor of K keypoints
    ) -> torch.Tensor:            
            _heatmaps = heatmaps.clone().detach() #TODO: check if clone is needed or if detach suffices
            _coords = coords.clone().detach()
            heatmap_height = _heatmaps.shape[-2]
            heatmap_width = _heatmaps.shape[-1]
            if self.grid_type != 'coord':
                _coords[..., 0] = (_coords[..., 0] + 1.0) / 2.0 * heatmap_width if self.grid_type == 'ndc' else _coords[..., 0] * heatmap_width
                _coords[..., 1] = (_coords[..., 1] + 1.0) / 2.0 * heatmap_height if self.grid_type == 'ndc' else _coords[..., 1] * heatmap_height            
            _coords = self._taylor(_coords, _heatmaps)
            if self.grid_type == 'ndc':
                _coords = self._convert_ndc(_coords, heatmap_height, heatmap_width)
            if self.grid_type == 'norm':
                _coords = self._convert_norm(_coords, heatmap_height, heatmap_width)
            return torch.flip(_coords, dims=[-1]) if self.flip else _coords


from moai.monads.utils.common import (
    expand_dims,
    dim_list,
    dims
)

import torch
import functools
import typing

#TODO: use torch.flatten?
def flatten_spatial_dims(
    tensor:         torch.Tensor,
    spatial_start_index:    int=2,
) -> torch.Tensor:
    dims = [*tensor.shape[:spatial_start_index]] + [-1]
    return tensor.view(*dims)

spatial_dim_list = functools.partial(dim_list, start_index=2)
expand_spatial_dims = functools.partial(expand_dims, start_index=2)
spatial_dims = functools.partial(dims, start_index=2)



from moai.supervision.losses.regression.L1 import L1
from moai.supervision.losses.regression.L2 import L2
from moai.supervision.losses.regression.cosine_distance import CosineDistance

__all__ = [
    "CosineDistance",
    "L1",
    "L2",
]


import toolz
import typing
import torch
import torchvision
import cv2
import numpy
import os

#NOTE: Check if needed for opencv versions over 4.5.5.62
if cv2.getVersionMajor() >= 4 and cv2.getVersionMinor() >= 5:
    os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"

def load_color(filename: str, **kwargs) -> torch.Tensor:    
    if 'position' in kwargs:
        filename = filename.replace('center', kwargs['position'])
    key = 'color' if 'position' not in kwargs else f'color_{kwargs["position"]}'
    return {key: torchvision.io.read_image(filename) / 255.0 }

def load_depth(filename: str, max_depth: float=8.0, **kwargs) -> torch.Tensor:
    if 'position' in kwargs:
        filename = filename.replace('center', kwargs['position'])
    depth_filename = filename.replace('emission', 'depth').replace('.png', '.exr')
    if 'filmic' in depth_filename.split(os.sep)[-3]:
        depth_filename = depth_filename.replace('_filmic', '')
    depth = torch.from_numpy(
        cv2.imread(depth_filename, cv2.IMREAD_ANYDEPTH)
    ).unsqueeze(0)
    #NOTE: add a micro meter to allow for thresholding to extact the valid mask
    depth[depth > max_depth] = max_depth + 1e-6
    key = 'depth' if 'position' not in kwargs else f'depth_{kwargs["position"]}'
    return {
        key: depth
    }

def load_normal(filename: str, **kwargs) -> torch.Tensor:
    if 'position' in kwargs:
        filename = filename.replace('center', kwargs['position'])
    normal_filename = filename.replace('emission', 'normal_map').replace('.png', '.exr')
    if 'filmic' in normal_filename.split(os.sep)[-3]:
        normal_filename = normal_filename.replace('_filmic', '')
    key = 'normal' if 'position' not in kwargs else f'normal_{kwargs["position"]}'
    return {
        key: torch.from_numpy(cv2.imread(normal_filename, 
            cv2.IMREAD_ANYDEPTH | cv2.IMREAD_ANYCOLOR
        ).transpose(2, 0, 1))
    }

def load_semantic(filename: str, **kwargs) -> torch.Tensor:
    if 'position' in kwargs:
        filename = filename.replace('center', kwargs['position'])
    key = 'semantic' if 'position' not in kwargs else f'semantic_{kwargs["position"]}'
    return {
        key: torch.from_numpy(cv2.imread(
            filename.replace('emission', 'semantic_map').replace('.png', '.exr')
        )).long()[:,:,0]
    }

def load_structure(filename: str, **kwargs) -> torch.Tensor:
    if 'position' in kwargs:
        filename = filename.replace('center', kwargs['position'])
    key = 'structure' if 'position' not in kwargs else f'structure_{kwargs["position"]}'
    return { 
        key: torch.from_numpy(cv2.imread(
            filename.replace('emission', 'layout').replace('.png', '.exr')
        )).long()
    }

def load_layout(filename: str, **kwargs) -> typing.Mapping[str, torch.Tensor]:
    if 'position' in kwargs:
        filename = filename.replace('center', kwargs['position'])
    ret = { }
    with numpy.load(filename) as f:
        layout = f['arr_0.npy']    
    ret["top_layout"] = torch.from_numpy(layout[0, ...])
    ret["top_weights"] = (ret["top_layout"] > 0).float()
    ret["top_layout"] = (128. - ret["top_layout"]) / 128.
    ret["bottom_layout"] = torch.from_numpy(layout[1, ...])
    ret["bottom_weights"] = (ret["bottom_layout"] > 0).float()
    ret["bottom_layout"] = (ret["bottom_layout"] - 128.) / 128.
    ret["top_layout"] *= ret["top_weights"]
    ret["bottom_layout"] *= ret["bottom_weights"]
    if 'position' in kwargs:
        ret = toolz.keymap(lambda k: f'{k}_{kwargs["position"]}', ret)
    return ret



import typing
import logging

def assert_numeric(
    logger:             logging.Logger,
    name:               str,
    value:              typing.Union[float, int],
    min_value:          typing.Union[float, int]=None,  
    max_value:          typing.Union[float, int]=None,
) -> None:
    if min_value is not None and max_value is not None\
        and (value < min_value or value > max_value):
            logger.error(f"Parameter {name} (value: {value}) should be in range [{min_value}, {max_value}].")
    if min_value is not None and max_value is None and value < min_value:
            logger.error(f"Parameter {name} (value: {value}) should be larger than {min_value}.")
    if max_value is not None and min_value is None and value > max_value:
            logger.error(f"Parameter {name} (value: {value}) should be smaller than {max_value}.")

def assert_non_negative(
    logger:             logging.Logger,
    name:               str,
    value:              typing.Union[float, int],
):
    if value is not None and value < 0.0:
        logger.error(f"Parameter {name} (value: {value}) should not be negative.")

def assert_negative(
    logger:             logging.Logger,
    name:               str,
    value:              typing.Union[float, int],
):
    if value is not None and value >= 0.0:
        logger.error(f"Parameter {name} (value: {value}) should be negative.")

def assert_positive(
    logger:             logging.Logger,
    name:               str,
    value:              typing.Union[float, int],
):
    if value is not None and value <= 0.0:
        logger.error(f"Parameter {name} (value: {value}) should be positive.")



import torch

#NOTE: from https://arxiv.org/pdf/2006.08195.pdf 
#TODO: specific parameter/weight init func

__all__ = ['Snake']

class Snake(torch.nn.Module):
    def __init__(self,
        alpha: float=1.0,
    ):
        super(Snake, self).__init__()
        self.alpha = alpha
        self.one_over_alpha = 1.0 / alpha

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        s = torch.sin(self.alpha * x)
        return x + self.one_over_alpha * (s ** 2)


import torch
import pysm
import re
import toolz
import functools
import itertools
import inspect
import typing
import logging
import omegaconf.omegaconf

log = logging.getLogger(__name__)

__all__ = [
    'get_result_fillers',
    'resolve_io_config',
    'split_as',
    'force_list',
]

def _tuple_add(
    tensor_dict: typing.Dict[str, torch.Tensor],
    tensor: torch.Tensor,
    key: str
) -> None:
    tensor_dict[key] = tensor

def _list_add(
    tensor_dict: typing.Dict[str, torch.Tensor],
    tensor_list: typing.List[torch.Tensor],
    key: str
) -> None:
    for i, tensor in enumerate(tensor_list):
        tensor_dict[key + "_" + str(i)] = tensor

def _tensor_add(
    tensor_dict: typing.Dict[str, torch.Tensor], 
    tensor: torch.Tensor,
    key: str
) -> None:
    tensor_dict[key] = tensor

def _no_add(
    tensor_dict: typing.Dict[str, torch.Tensor],
    tensor: torch.Tensor
) -> None:
    pass

__ACTION_MAP__ = {
    "Tuple": _tuple_add,
    "List": _list_add,
    "Tensor": _tensor_add,
    "NoOp": _no_add
}

class ReturnTypeParser(object):
    def __init__(self, keys):
        self.keys = keys
        self.counter = 0
        self.sm = self._create_state_machine()
        self.result = []


    def _create_state_machine(self) -> pysm.StateMachine:
        sm = pysm.StateMachine('sm')

        initial = pysm.State('Initial')
        move_to = pysm.State('MoveTo')
        advance = pysm.State('Advance')
        begin = pysm.State('Begin')
        end = pysm.State('End')
        tensor = pysm.State('Tensor')
        sequence = pysm.State('Sequence')

        sm.add_state(initial, initial=True)
        sm.add_state(move_to)
        sm.add_state(advance)
        sm.add_state(begin)
        sm.add_state(end)
        sm.add_state(end)
        sm.add_state(tensor)
        sm.add_state(sequence)

        sm.add_transition(initial, move_to,
            events=['parse'], input=['Tuple', 'List'],
            action=self._set_func)
        sm.add_transition(initial, tensor,
            events=['parse'], input=['Tensor'],
            action=self._single_out)
        #TODO: from list to tuple before end should generate an error !
        sm.add_transition(move_to, begin, 
            events=['parse'], input='[',
            action=self._start)
        sm.add_transition(begin, tensor,
            events=['parse'], input=['Tensor'],
            action=self._register)
        sm.add_transition(advance, tensor,
            events=['parse'], input=['Tensor'],
            action=self._register)
        sm.add_transition(tensor, advance,
            events=['parse'], input=',',
            action=self._advance)
        sm.add_transition(end, advance,
            events=['parse'], input=',',
            action=self._advance)
        sm.add_transition(advance, move_to,
            events=['parse'], input=['Tuple', 'List'],
            action=self._set_func)
        sm.add_transition(tensor, end,
            events=['parse'], input=']',
            action=self._end)
        
        sm.initialize()
        return sm

    def parse(self, tokens: str) -> typing.List[typing.Callable]:
        for token in tokens:
            self.sm.dispatch(pysm.Event('parse', input=token))
        return self.result
        
    def _list_only(self, state: pysm.State, event: pysm.Event) -> None:
        key = self.keys[self.counter]
        if not key:
            self.result.append(
                __ACTION_MAP__['NoOp']
            )
        else:
            self.result.append(functools.partial(
                __ACTION_MAP__['Tuple'], key=key
            ))

    def _single_out(self, state: pysm.State, event: pysm.Event) -> None:
        key = self.keys[self.counter]
        if not key:
            self.result.append(
                __ACTION_MAP__['NoOp']
            )
        else:
            self.result.append(functools.partial(
                __ACTION_MAP__['Tensor'], key=key
            ))

    def _end(self, state: pysm.State, event: pysm.Event) -> None:
        self.sm.stack.pop()

    def _advance(self, state: pysm.State, event: pysm.Event) -> None:
        self.counter += 1

    def _register(self, state: pysm.State, event: pysm.Event) -> None:
        # assert self.counter < len(self.keys),\
        # NOTE: '''
        #     Output tensor naming parsing stopped as an insufficient number of keys are present.
        #     Please check the output return type annotation and provide keys for all resulting tensors. 
        #     * Ignored tensor results can be omitted by using an empty string
        # '''
        key = self.keys[self.counter] if self.counter < len(self.keys) else ''
        if not key:
            if self.counter >= len(self.keys):
                log.warning(f"Less output keys were given ({self.keys}) than the operation supports."
                "This can either be due to an error or to ignore some outputs as they are not used."
                "In the latter case, the warning can be supressed using an empty string key \'\'.")
            self.result.append(
                __ACTION_MAP__['NoOp']
            )
        else:
            self.result.append(functools.partial(
                self.sm.stack.peek(), key=key
            ))

    def _start(self, state: pysm.State, event: pysm.Event) -> None:
        pass

    def _set_func(self, state: pysm.State, event: pysm.Event) -> None:
        self.sm.stack.push(__ACTION_MAP__[event.input])


def get_result_fillers(
    module: torch.nn.Module,
    keys: typing.Sequence[str]
):
    hints = str(typing.get_type_hints(module.forward)['return']) # return type annotations
    hints = hints.replace('typing.', '')\
        .replace('torch.','') # remove module names to keep only the pure type names
    tokens = re.split(r'(\W)', hints) # split in words
    tokens = list(toolz.remove(lambda s: not s or s.isspace(), tokens)) # remove empties
    rtp = ReturnTypeParser(keys=keys)
    fillers = rtp.parse(tokens) # get type specific filler functors
    
    def filler(
        tensor_dict: typing.Dict[str, torch.Tensor],
        out_tensor: torch.Tensor,
        fillers: typing.List[typing.Callable]
    ): # actual filler functor execution
        #TODO: check if out_tensor is pure tensor, then zip unwraps its batch dimension?
        for item, visitor in zip(
            [out_tensor] if (type(out_tensor) is list or type(out_tensor) is torch.Tensor) else out_tensor,
            itertools.cycle(fillers)
        ):
            visitor(tensor_dict, item)

    return functools.partial(filler, fillers=fillers) # return lambda with cached keys

def resolve_io_config(
    cfg: typing.Union[str, omegaconf.ListConfig]
):
    if type(cfg) is str:
        return [[cfg]]
    elif type(cfg) is omegaconf.ListConfig:
        return cfg if all(type(i) is omegaconf.ListConfig for i in cfg) else [cfg]

def split_as(
    frm: typing.List[typing.Any], 
    to: typing.List[typing.Any]
):
    if len(frm) == len(to):
        return frm
    elif len(frm) < len(to) and len(frm[0]) == len(to):
        return [[x] for x in frm[0]]

def force_list(i: typing.Union[str, typing.Any]):
    return [i] if type(i) is str else i


import torch
import typing

__all__ = ['PID']

#NOTE: from https://github.com/jettify/pytorch-optimizer/blob/master/torch_optimizer/pid.py

class PID(torch.optim.Optimizer):
    r"""Implements PID optimization algorithm.
    
    - **Paper**: [A PID Controller Approach for Stochastic Optimization of Deep Networks](https://csjunxu.github.io/paper/TNNLS2020.PDF)
    - **Implementation**: [GitHub @ jettify](https://github.com/jettify/pytorch-optimizer)    

    Arguments:
        params: iterable of parameters to optimize or dicts defining
            parameter groups
        lr: learning rate (default: 1e-3)
        momentum: momentum factor (default: 0.0)
        weight_decay: weight decay (L2 penalty) (default: 0.0)
        dampening: dampening for momentum (default: 0.0)
        derivative: D part of the PID (default: 10.0)
        integral: I part of the PID (default: 5.0)

    Example:
        >>> import torch_optimizer as optim
        >>> optimizer = optim.PID(model.parameters(), lr=0.001, momentum=0.1)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ http://www4.comp.polyu.edu.hk/~cslzhang/paper/CVPR18_PID.pdf

    Note:
        Reference code: https://github.com/tensorboy/PIDOptimizer
    """

    def __init__(self,
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        momentum: float=0.0,
        dampening: float=0,
        weight_decay: float=0.0,
        integral: float=5.0,
        derivative: float=10.0,
    ) -> None:
        defaults = dict(
            lr=lr,
            momentum=momentum,
            dampening=dampening,
            weight_decay=weight_decay,
            integral=integral,
            derivative=derivative,
        )
        if lr <= 0.0:
            raise ValueError('Invalid learning rate: {}'.format(lr))
        if momentum < 0.0:
            raise ValueError('Invalid momentum value: {}'.format(momentum))
        if weight_decay < 0.0:
            raise ValueError(
                'Invalid weight_decay value: {}'.format(weight_decay)
            )
        if integral < 0.0:
            raise ValueError('Invalid PID integral value: {}'.format(integral))
        if derivative < 0.0:
            raise ValueError(
                'Invalid PID derivative value: {}'.format(derivative)
            )

        super(PID, self).__init__(params, defaults)

    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            integral = group['integral']
            derivative = group['derivative']
            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = p.grad.data
                if weight_decay != 0:
                    d_p.add_(p.data, alpha=weight_decay)
                if momentum != 0:
                    param_state = self.state[p]
                    if 'i_buffer' not in param_state:
                        i_buf = param_state['i_buffer'] = torch.zeros_like(p)
                        i_buf.mul_(momentum).add_(d_p)
                    else:
                        i_buf = param_state['i_buffer']
                        i_buf.mul_(momentum).add_(d_p, alpha=1 - dampening)
                    if 'grad_buffer' not in param_state:
                        g_buf = param_state['grad_buffer'] = torch.zeros_like(
                            p
                        )
                        g_buf = d_p

                        d_buf = param_state['d_buffer'] = torch.zeros_like(p)
                        d_buf.mul_(momentum).add_(d_p - g_buf)
                    else:
                        d_buf = param_state['d_buffer']
                        g_buf = param_state['grad_buffer']
                        d_buf.mul_(momentum).add_(
                            d_p - g_buf, alpha=1 - momentum
                        )
                        self.state[p]['grad_buffer'] = d_p.clone()

                    d_p = d_p.add_(i_buf, alpha=integral).add_(
                        d_buf, alpha=derivative
                    )
                p.data.add_(d_p, alpha=-group['lr'])
        return loss


from moai.monads.utils.common import dim_list

import torch
import functools

class Delta(torch.nn.Module):
    def __init__(self, 
        threshold: float=1.25
    ):
        super(Delta, self).__init__()
        self.threshold = threshold

    def forward(self,
        gt:             torch.Tensor,
        pred:           torch.Tensor,
        weights:        torch.Tensor=None,
    ) -> torch.Tensor: #NOTE: no mean
        errors =  (torch.max((gt / pred), (pred / gt)) < self.threshold).float()
        if weights is None:
            return torch.mean(errors)
        else:
            return torch.mean(
                torch.sum(errors * weights, dim=dim_list(gt))
                / torch.sum(weights, dim=dim_list(gt))
            )

Delta1 = functools.partial(Delta, threshold=1.25)
Delta2 = functools.partial(Delta, threshold=1.25 ** 2)
Delta3 = functools.partial(Delta, threshold=1.25 ** 3)


import torch
import typing

__all__ = ["Intrinsics"]

def _intrinsics_projection_matrix(
    fx:     float,
    fy:     float,
    px:     float,
    py:     float,
) -> torch.Tensor:
    return torch.Tensor([
        [fx,    0.0,    px],
        [0.0,   fy,     py],
        [0.0,   0.0,    1.0],
    ]).float()

class Intrinsics(torch.nn.Module): #NOTE: check if it is transposed
    def __init__(self,
        focal:          typing.Optional[typing.Sequence[float]]=None,
        principal:      typing.Optional[typing.Sequence[float]]=None,
        fov:            float=64.69, # in degrees
        width:          int=320,
        height:         int=240,        
        mode:           str='fov' # one of 'intrinsics', 'fov'
    ):
        super(Intrinsics, self).__init__()
        if mode == 'fov':
            tanfov = torch.tan((torch.deg2rad(torch.tensor(fov) / 2.0).float()))
            aspect_ratio =  width / height
            fx, fy = -1.0 * width / 2.0 / tanfov, height / 2.0 / tanfov * aspect_ratio
            px, py = width / 2.0, height / 2.0
            intrinsics_matrix = _intrinsics_projection_matrix(fx, fy, px, py)
        elif mode == 'intrinsics':
            fx, fy = focal
            px, py = principal if principal is not None else (width / 2.0, height / 2.0)
            intrinsics_matrix = _intrinsics_projection_matrix(fx, fy, px, py)
        self.register_buffer("intrinsics_matrix", intrinsics_matrix)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b = x.shape[0]
        return self.intrinsics_matrix.expand(b, *self.intrinsics_matrix.shape[0:])


import torch

class Relative(torch.nn.Module):
    def __init__(self,
        
    ):
        super(Relative, self).__init__()

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return y.inverse() @ x # from x -> to y 


class Inverse(torch.nn.Module):
    def __init__(self,
        
    ):
        super(Inverse, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.inverse()

class Transpose(torch.nn.Module):
    def __init__(self,
        
    ):
        super(Transpose, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.transpose(x, -1, -2)


import torch
import typing

__all__ = ['Apollo', 'ApolloW']

#NOTE: from https://github.com/XuezheMax/apollo/blob/master/optim/apollo.py

class Apollo(torch.optim.Optimizer):
    r"""Implements Atom algorithm.

    - **Paper**: [Apollo: An Adaptive Parameter-wise Diagonal Quasi-Newton Method for Nonconvex Stochastic Optimization](https://arxiv.org/pdf/2009.13586.pdf)
    - **Implementation**: [GitHub @ XuezheMax](https://github.com/XuezheMax/apollo)

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): learning rate
        beta (float, optional): coefficient used for computing
            running averages of gradient (default: 0.9)
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-4)
        warmup (int, optional): number of warmup steps (default: 0)
        init_lr (float, optional): initial learning rate for warmup (default: 0.01)
        weight_decay (float, optional): weight decay coefficient (default: 0)

    """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        #lr: float=torch.optim.optimizer.required,
        lr=None,
        beta: float=0.9,
        eps: float=1e-4,
        warmup: int=100,
        init_lr: float=0.01,
        weight_decay: float=0
    ):
        if not 0.0 < lr:
            raise ValueError("Invalid learning rate value: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= beta < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(beta))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        if not 0.0 <= warmup:
            raise ValueError("Invalid warmup updates: {}".format(warmup))
        if not 0.0 <= init_lr <= 1.0:
            raise ValueError("Invalid initial learning rate: {}".format(init_lr))

        defaults = dict(lr=lr, beta=beta, eps=eps, warmup=warmup,
                        init_lr=init_lr, base_lr=lr, weight_decay=weight_decay)
        super(Apollo, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(Apollo, self).__setstate__(state)

    @torch.no_grad()
    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg_grad'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state['approx_hessian'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Previous update direction
                    state['update'] = torch.zeros_like(p, memory_format=torch.preserve_format)

                # Calculate current lr
                if state['step'] < group['warmup']:
                    curr_lr = (group['base_lr'] - group['init_lr']) * state['step'] / group['warmup'] + group['init_lr']
                else:
                    curr_lr = group['lr']

                # Perform optimization step
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError('Atom does not support sparse gradients.')

                # Perform step weight decay
                if group['weight_decay'] != 0:
                    grad = grad.add(p, alpha=group['weight_decay'])

                beta = group['beta']
                exp_avg_grad = state['exp_avg_grad']
                B = state['approx_hessian']
                d_p = state['update']

                state['step'] += 1
                bias_correction = 1 - beta ** state['step']
                alpha = (1 - beta) / bias_correction

                # Update the running average grad
                delta_grad = grad - exp_avg_grad
                exp_avg_grad.add_(delta_grad, alpha=alpha)

                denom = d_p.norm(p=4).add(group['eps'])
                d_p.div_(denom)
                v_sq = d_p.mul(d_p)
                delta = delta_grad.div_(denom).mul_(d_p).sum().mul(-alpha) - B.mul(v_sq).sum()

                # Update B
                B.addcmul_(v_sq, delta)

                # calc direction of parameter updates
                denom = B.abs().clamp_(min=1)
                d_p.copy_(exp_avg_grad.div(denom))

                p.add_(d_p, alpha=-curr_lr)

        return loss

class ApolloW(torch.optim.Optimizer):
    r"""Implements Atom algorithm.

        Arguments:
            params (iterable): iterable of parameters to optimize or dicts defining
                parameter groups
            lr (float): learning rate
            beta (float, optional): coefficient used for computing
                running averages of gradient (default: 0.9)
            eps (float, optional): term added to the denominator to improve
                numerical stability (default: 1e-4)
            warmup (int, optional): number of warmup steps (default: 0)
            init_lr (float, optional): initial learning rate for warmup (default: 0.01)
            weight_decay (float, optional): weight decay coefficient (default: 0)

        """
    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        #lr: float=torch.optim.optimizer.required,
        lr=None,
        beta: float=0.9,
        eps: float=1e-4,
        warmup: int=100,
        init_lr: float=0.01,
        weight_decay: float=0
    ):
        if not 0.0 < lr:
            raise ValueError("Invalid learning rate value: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= beta < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(beta))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        if not 0.0 <= warmup:
            raise ValueError("Invalid warmup updates: {}".format(warmup))
        if not 0.0 <= init_lr <= 1.0:
            raise ValueError("Invalid initial learning rate: {}".format(init_lr))

        defaults = dict(lr=lr, beta=beta, eps=eps, warmup=warmup,
                        init_lr=init_lr, base_lr=lr, weight_decay=weight_decay)
        super(ApolloW, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(ApolloW, self).__setstate__(state)

    @torch.no_grad()
    def step(self, closure: typing.Optional[typing.Callable[[], float]]=None) -> typing.Optional[float]:
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg_grad'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state['approx_hessian'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Previous update direction
                    state['update'] = torch.zeros_like(p, memory_format=torch.preserve_format)

                # Calculate current lr
                if state['step'] < group['warmup']:
                    curr_lr = (group['base_lr'] - group['init_lr']) * state['step'] / group['warmup'] + group['init_lr']
                else:
                    curr_lr = group['lr']

                # Perform optimization step
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError('Atom does not support sparse gradients.')

                beta = group['beta']
                exp_avg_grad = state['exp_avg_grad']
                B = state['approx_hessian']
                d_p = state['update']

                state['step'] += 1
                bias_correction = 1 - beta ** state['step']
                alpha = (1 - beta) / bias_correction

                # Update the running average grad
                delta_grad = grad - exp_avg_grad
                exp_avg_grad.add_(delta_grad, alpha=alpha)

                denom = d_p.norm(p=4).add(group['eps'])
                d_p.div_(denom)
                v_sq = d_p.mul(d_p)
                delta = delta_grad.div_(denom).mul_(d_p).sum().mul(-alpha) - B.mul(v_sq).sum()

                # Update B
                B.addcmul_(v_sq, delta)

                # calc direction of parameter updates
                denom = B.abs().clamp_(min=1)
                d_p.copy_(exp_avg_grad.div(denom))

                # Perform step weight decay
                if group['weight_decay'] != 0:
                    d_p.add_(p, alpha=group['weight_decay'])

                p.add_(d_p, alpha=-curr_lr)

        return loss


import torch
import typing
import logging
import sys
import functools

__all__ = ["Empty"]

log = logging.getLogger(__name__)

class Empty(torch.utils.data.Dataset):
    def __init__(self):
        log.warning("Constructing an empty dataset !")
    
    def __len__(self) -> int:
        return sys.maxsize

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        return {}

class Dummy(torch.utils.data.Dataset):
    def __init__(self,
        size:       int=sys.maxsize,
        **kwargs: typing.Mapping[str, typing.Any],
    ):
        self.generators = {}
        self.size = size        
        for key, props in kwargs.items():
            self.generators[key] = functools.partial(getattr(torch, props.init or 'rand'),
                size=tuple(props.shape),
            )
        log.info(f"Using a custom dataset producing {len(kwargs)} dummy tensors ([{list(kwargs.keys())}]) of size {len(self)}.")
    
    def __len__(self) -> int:
        return self.size

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        return dict((k, g()) for k, g in self.generators.items())


import moai.networks.lightning as minet
import moai.utils.parsing.rtp as mirtp

import torch
import inspect
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging 

__all__ = ['Autoencoder']

log = logging.getLogger(__name__)

class Autoencoder(minet.FeedForward):
    def __init__(self, 
        configuration:  omegaconf.DictConfig, 
        modules:        omegaconf.DictConfig,        
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,
        monads:         omegaconf.DictConfig=None,
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(Autoencoder, self).__init__(
            data=data, parameters=parameters,
            feedforward=feedforward, monads=monads,
            supervision=supervision, validation=validation,
            visualization=visualization, export=export,
        ) 
        self.encoder = hyu.instantiate(modules['encoder'])
        self.decoder = hyu.instantiate(modules['decoder'])
        self.enc_fwds, self.dec_fwds = [], []

        params = inspect.signature(self.encoder.forward).parameters
        enc_in = list(zip(*[mirtp.force_list(configuration.encoder[prop]) for prop in params]))
        enc_out = mirtp.split_as(mirtp.resolve_io_config(configuration.encoder['out']), enc_in)
        
        self.enc_res_fill = [mirtp.get_result_fillers(self.encoder, out) for out in enc_out]        
        get_enc_filler = iter(self.enc_res_fill)
        
        for keys in enc_in:
            self.enc_fwds.append(lambda td,
                tk=keys,
                args=params.keys(),
                enc=self.encoder,
                filler=next(get_enc_filler):
                    filler(td, enc(**dict(zip(args,
                        list(
                                td[k] if type(k) is str
                                else list(td[j] for j in k)
                            for k in tk
                        )
                    ))))
            )
        
        params = inspect.signature(self.decoder.forward).parameters
        dec_in = list(zip(*[mirtp.force_list(configuration.decoder[prop]) for prop in params]))
        dec_out = mirtp.split_as(mirtp.resolve_io_config(configuration.decoder['out']), dec_in)

        self.dec_res_fill = [mirtp.get_result_fillers(self.decoder, out) for out in dec_out]
        get_dec_filler = iter(self.dec_res_fill)
        
        for keys in dec_in:
            self.dec_fwds.append(lambda td,
                tk=keys,
                args=params.keys(), 
                dec=self.decoder,
                filler=next(get_dec_filler):
                    filler(td, dec(**dict(zip(args,
                        list(
                                td[k] if type(k) is str
                                else list(td[j] for j in k)
                            for k in tk
                        )
                    ))))
            )

    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        for e, d in zip(self.enc_fwds, self.dec_fwds):
            e(td)
            d(td)
        return td


from moai.utils.engine.collection import Collection
from moai.utils.engine.single import Single #NOTE: order matters as single inherits from collection, can switch to non package import to fix
from moai.utils.engine.interval import Interval
from moai.utils.engine.noop import NoOp #NOTE: order matters as noop inherits from interval, can switch to non package import to fix

__all__ = [
    "NoOp",
    "Collection",
    "Single",
    "Interval",
]


from moai.parameters.selectors.named import NamedParameterSelector
from moai.parameters.selectors.monad import MonadParameterSelector
from moai.parameters.selectors.model import ModelParameterSelector
from moai.parameters.selectors.module import ModuleParameterSelector

__all__ = [
    'NamedParameterSelector',
    'MonadParameterSelector',
    'ModelParameterSelector',
    'ModuleParameterSelector',
]


from moai.visualization.visdom.base import Base
from moai.utils.color.colorize import get_colormap, COLORMAPS

import torch
import visdom
import functools
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["Feature2d"]

class Feature2d(Base):
    def __init__(self,
        image:           typing.Union[str, typing.Sequence[str]],
        type:          typing.Union[str, typing.Sequence[str]],
        colormap:      typing.Union[str, typing.Sequence[str]],
        transform:     typing.Union[str, typing.Sequence[str]],
        name:           str="default",
        ip:             str="http://localhost",
        port:           int=8097,   
    ):
        super(Feature2d, self).__init__(name, ip, port)
        self.images = [image] if isinstance(image, str) else list(image)
        self.types = [type] if isinstance(type, str) else list(type)
        self.transforms = [transform] if isinstance(transform, str) else list(transform)
        self.colormaps = [colormap] if isinstance(colormap, str) is str else list(colormap)
        self.viz_map = {
            'color': functools.partial(self.__viz_color, self.visualizer),
            'heatmap': functools.partial(self.__viz_heatmap, self.visualizer),
        }
        self.transform_map = {
            'none': functools.partial(self.__no_transform),
            'minmax': functools.partial(self.__minmax_normalization)
        }
        self.colorize_map = { "none": lambda x: x }
        self.colorize_map.update(COLORMAPS)

    @property
    def name(self) -> str:
        return self.env_name
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for k, t, tf, c in zip(self.images, self.types, self.transforms, self.colormaps):
            if not k:
                continue
            tensor = tensors[k]
            b, h, _, __ = tensor.size()            
            for i in range(h):
                key = k + "_{}".format(i)
                self.viz_map[t](
                    self.colorize_map[c](
                        self.transform_map[tf](
                            tensor[:, i, ...].unsqueeze(1)
                        )), key, key, self.name
                )

    @staticmethod
    def __viz_color(
        visdom: visdom.Visdom,
        tensor: torch.Tensor,
        key: str,
        win: str,
        env: str
    ) -> None:
        visdom.images(
            tensor,
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': 50,
            }
        )

    @staticmethod
    def __viz_heatmap(
        visdom: visdom.Visdom,
        tensor: torch.Tensor,
        key: str,
        win: str,
        env: str
    ) -> None:
        b, _, __, ___ = tensor.size() # assumes [B, C, H, W], i.e. 2d train
        heatmaps = torch.flip(tensor, dims=[2]).detach().cpu()
        for i in range(b):
            opts = (
            {
                'title': key + "_{}".format(i),
                'colormap': 'Viridis'
            })
            visdom.heatmap(heatmaps[i, :, :, :].squeeze(), opts=opts, win=win + str(i))

    @staticmethod #TODO: refactor these into a common module
    def __no_transform(tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    @staticmethod #TODO: refactor these into a common module
    def __minmax_normalization(tensor: torch.Tensor) -> torch.Tensor:
        b, c, __, ___ = tensor.size()
        min_v = torch.min(tensor.view(b, c, -1), dim=2, keepdim=True)[0].unsqueeze(3)
        max_v = torch.max(tensor.view(b, c, -1), dim=2, keepdim=True)[0].unsqueeze(3)
        return (tensor - min_v) / (max_v - min_v)
    


